""" 
# Datatypes Decorator 

Wraps `@pydantic.dataclasses.dataclass` so that we can make forward type-references 
throughout the package, and then sort them all out at the end of import-time. 

The intended usage is such that `__init__.py` includes

```python
# Import all the other code 
from .other_module1 import Whatever
from .other_module2 import WhateverElse

# Then at the end, get and run the forward-reference updater-method
from .datatype import _update_forward_refs

_update_forward_refs()
```

Notes: 
- `@datatype` is designed solely to work *intra-package*. 
  - Attempts to import and use it after `__init__.py` completes will generally fail. 
  - Importing this function into other packages is therefore highly discouraged. (Copying it is quite easy though.) 
- `@datatype` only works on modules which are imported before `_update_forward_refs()` is run. 
  - Generally this means modules which are imported as part of `__init__.py`
- `@datatype` is designed solely to work on `pydantic.dataclasses.dataclass`es. 
  - Notable exceptions include *union types* thereof, which do not have the necessary fields/ methods. 
"""

from typing import TypeVar, Type, Optional
from pydantic import __version__ as _pydantic_version


_pydantic_major_version = int(_pydantic_version.split(".")[0])
if _pydantic_major_version > 2 or _pydantic_major_version < 1:
    msg = "Error reading Pydantic version. Should be either 1.x or 2.x."
    raise ImportError(msg)

if _pydantic_major_version == 1:
    from pydantic import Extra, BaseModel
    from pydantic.json import pydantic_encoder as pydantic_json_encoder

    PYDANTIC_V2 = False

    class OurBaseConfig:
        allow_extra = Extra.forbid

    class AllowArbConfig(OurBaseConfig):
        arbitrary_types_allowed = True

    def _update_forward_refs():
        """Update all the forward type-references"""
        for tp in datatypes:
            tp.__pydantic_model__.update_forward_refs()

else:  # _pydantic_major_version==2
    from pydantic import Extra, BaseModel, RootModel, BeforeValidator
    from pydantic.deprecated.json import pydantic_encoder as pydantic_json_encoder

    PYDANTIC_V2 = True
    OurBaseConfig = dict(allow_extra="forbid", validate_default=True)
    AllowArbConfig = dict(
        allow_extra="forbid", validate_default=True, arbitrary_types_allowed=True
    )

    def _update_forward_refs():
        """Update all the forward type-references"""
        ...
        # for tp in datatypes:
        #     tp.model_rebuild()


from pydantic.dataclasses import dataclass

T = TypeVar("T")
datatypes = []  # The list of defined datatypes


def _datatype(cls: Type[T], *, config: Optional[Type] = None, **kwargs) -> Type[T]:
    """# Inner implementation of `@datatype`"""

    # Get the default `Config` if none is provided
    config = config or OurBaseConfig

    # Convert `cls` to a `pydantic.dataclasses.dataclass`,
    cls = dataclass(cls, config=config, **kwargs)

    # And add it to the list of datatypes
    datatypes.append(cls)
    return cls


def datatype(cls: Optional[Type[T]] = None, **kwargs) -> Type[T]:
    """Register a class as a datatype."""

    # NOTE: the return type here is really, like,
    # `Union[Type[T], Callable[[Type[T]], Type[T]]`
    # But do ya really want that, or just to know it works as a decorator.

    inner = lambda c: _datatype(c, **kwargs)
    if cls is None:
        return inner  # Called with parens, e.g. `@datatype()`
    return inner(cls)  # Called without parens



# Std-Lib Imports
from typing import Optional, Set

# Local imports
from . import attrmagic
from .connect import connectable
from .sliceable import sliceable
from .concatable import concatable, is_concatable


@attrmagic.only_set_known_attrs
@attrmagic.init
@sliceable
@concatable
@connectable
class Concat:
    """
    # Signal Concat(enations)
    "Sequences" of Signals and other Connectable objects.
    Uses *Python-convention* ordering, in which "LSBs", i.e. index 0, are specified first.
    """

    def __init__(self, *parts):
        invalid_parts = [p for p in parts if not is_concatable(p)]
        if invalid_parts:
            raise TypeError(f"Invalid `Concat` of {invalid_parts}")

        for part in parts:
            part._concats.add(self)

        # Store the `parts` to a tuple
        self.parts = tuple(parts)

        # Inner management data
        self._connected_ports: Set["PortRef"] = set()
        self._width: Optional[int] = None
        self._slices: Set["Slice"] = set()
        self._concats: Set["Concat"] = set()
        self._initialized = True

    def __eq__(self, other) -> bool:
        # Identity is equality
        return other is self

    def __hash__(self) -> bool:
        # Identity is equality
        return hash(id(self))

    def __repr__(self):
        return f"Concat(parts={len(self.parts)})"

    @property
    def width(self):
        # FIXME: whether to keep me around as a property
        # As is the elaboration mechanics do not use this property, but users may.
        from .elab.helpers.width import width

        return width(self)


__doc__ = Concat.__doc__
__all__ = ["Concat"]



"""
# External Module

Wrapper for circuits defined outside Hdl21. 
"""

from typing import Any, Optional, List, Type, Dict
from pydantic.dataclasses import dataclass

# VLSIR Imports
from vlsirtools import SpiceType

# Local imports
from .datatype import datatype
from .default import Default
from .call import param_call
from .source_info import source_info, SourceInfo
from .params import HasNoParams, isparamclass, _unique_name
from .signal import Signal, Visibility
from .instance import calls_instantiate


@datatype
class ExternalModule:
    """
    # External Module

    Wrapper for circuits defined outside Hdl21, such as:
    * Inclusion of existing SPICE or Verilog netlists
    * Foundry or technology-specific primitives

    Unlike `Modules`, `ExternalModules` include parameters to support legacy HDLs.
    Said parameters may only take on a limited number of datatypes, and may not be nested.
    Each `ExternalModule` stores a parameter-type field `paramtype`.
    Parameter types may be either `hdl21.paramclass`es or the built-in `dict`.
    Parameter-values are checked to be instances of `paramtype` at creation time.
    """

    name: str  # Module name. Used *directly* when exporting.
    port_list: List[Signal]  # Ordered Ports
    paramtype: Type[object] = HasNoParams  # Parameter-type `paramclass`
    desc: Optional[str] = None  # Description
    domain: Optional[str] = None  # Domain name, for references upon export
    spicetype: SpiceType = SpiceType.SUBCKT  # Spice type, for SPICE export

    @property
    def ports(self) -> dict:
        """Port dictionary, from name to `Signal`."""
        return {p.name: p for p in self.port_list}

    @property
    def Params(self) -> Type:
        """Type-style alias for the parameter-type."""
        return self.paramtype

    def __post_init__(self):
        """Post-Constructor Checks"""

        # Check for a valid parameter-type
        if not isparamclass(self.paramtype) and self.paramtype not in (dict, Dict):
            msg = f"Invalid parameter type {self.paramtype} for {self}. "
            msg += "Param types must be either `@paramclass`es or `dict`."
            raise ValueError(msg)

        # Internal tracking data: defining module/import-path
        self._source_info: Optional[SourceInfo] = source_info(get_pymodule=True)
        self._importpath = None

        # Now do some more checks on values
        for p in self.port_list:
            if not p.name:
                raise ValueError(f"Unnamed Primitive Port {p} for {self.name}")
            if p.vis != Visibility.PORT:
                msg = f"Invalid Primitive Port {p.name} on {self.name}; must have PORT visibility"
                raise ValueError(msg)

    def __call__(self, arg: Any = Default, **kwargs) -> "ExternalModuleCall":
        """Call to set an `ExternalModule`'s parameters.
        Returns an `ExternalModuleCall` combining the `ExternalModule` and parameter values.
        """
        params = param_call(callee=self, arg=arg, **kwargs)
        return ExternalModuleCall(module=self, params=params)

    def __eq__(self, other) -> bool:
        # Identity is equality
        return other is self

    def __hash__(self) -> bool:
        # Identity is equality
        return hash(id(self))


@calls_instantiate
@dataclass
class ExternalModuleCall:
    """# External Module Call
    A combination of an `ExternalModule` and its Parameter-values, typically generated by calling the `ExternalModule`.
    """

    module: ExternalModule
    params: Any

    def __post_init__(self):
        # Type-validate our parameters
        if not isinstance(self.params, self.module.paramtype):
            msg = f"Invalid parameter type {type(self.params)} for ExternalModule {self.module.name}. Must be {self.module.paramtype}"
            raise TypeError(msg)

    @property
    def name(self) -> str:
        return self.module.name + "(" + _unique_name(self.params) + ")"

    @property
    def ports(self) -> Dict[str, Signal]:
        return self.module.ports

    def __eq__(self, other) -> bool:
        """Call equality requires:
        * *Identity* between modules, and
        * *Equality* between parameter-values."""
        return self.module is other.module and self.params == other.params

    def __hash__(self):
        """Generator-Call hashing, consistent with `__eq__` above, uses:
        * *Identity* of its module, and
        * *Value* of its parameters.
        The two are joined for hashing as a two-element tuple."""
        return hash((id(self.module), self.params))


__all__ = ["ExternalModule", "ExternalModuleCall"]



"""
# Visibility 

Annotations for HDL objects to be visible inside or outside their parent, generally Module. 
Prominent examples include the annotations of which `Signal`s are exposed as ports. 
"""

from enum import Enum


class Visibility(Enum):
    """# Visibility Enumeration"""

    INTERNAL = "INTERNAL"  # Internal/ private
    PORT = "PORT"  # Exposed as a Port



"""
Spice-Class Simulation Interface 
"""

from textwrap import dedent
from warnings import warn
from enum import Enum
from typing import Union, Any, Optional, List
from pathlib import Path
from dataclasses import field

# Create a few aliases to the VLSIR sim-results types
# Note many of these are "re-exports", not used here, but convenient for use cases.
# Note `vsp.sim_data.SimResult` - the numpy based one - gets the `SimResult` name here.
from vlsirtools.spice.sim_data import AnalysisType, SimResult
from vlsirtools.spice import (
    SimOptions,
    SimResultUnion,
    ResultFormat,
    SupportedSimulators,
    sim as _vlsirtools_sim,
)
from vlsir.spice_pb2 import SimResult as SimResultProto

# Local Imports
from ..datatype import datatype, AllowArbConfig
from ..one_or_more import OneOrMore
from ..instance import Instance
from ..signal import Signal, Port
from ..instantiable import Instantiable, Module, ExternalModuleCall
from ..scalar import Scalar
from ..literal import Literal

# FIXME: deprecate `ParamVal`.
# Users may be importing it as `ParamVal` from here, so leave this name for now.
ParamVal = Scalar


def tb(name: str) -> Module:
    """Create a new testbench module.
    Includes the "testbench interface":
    a single, scalar, undirected Port named "VSS"."""
    tb = Module(name=name)
    tb.VSS = Port(width=1)
    return tb


def is_tb(m: Instantiable) -> bool:
    """Boolean indication of whether Instantiable `m` meets the test-bench interface."""
    if not isinstance(m, (Module, ExternalModuleCall)):
        # Also filter out `Primitive`s, which don't work as testbenches
        raise TypeError(f"Invalid un-instantiable argument {i} to `is_tb`")

    if len(m.ports) != 1:
        return False

    # There's exactly one port. Retrieve it from the `ports` dict.
    port = list(m.ports.values())[0]

    # While that port is *conventionally* called "VSS", it *can* be called anything.
    # The testbench interface is met so long as we have a single, scalar port.
    return port.width == 1


_simattrs = {}


def simattr(cls) -> type:
    """
    Add a class to the `simattrs` set of generation methods.
    This is the magic that enables methods including:
    ```python
    sim = Sim()
    sim.param(**kwargs)
    sim.dc(**kwargs)
    sim.tran(**kwargs)
    ```
    To create the classes defined here, and add them to `sim`.
    """

    # Class names are lower-cases to create the dictionary key,
    # which also becomes the method-name on `sim`.
    _simattrs[cls.__name__.lower()] = cls
    return cls


# Apply that "decorator" to the `Literal` class
simattr(Literal)


@simattr
@datatype
class Param:
    """Simulation Parameter-Value"""

    # Parameter Value
    val: Scalar
    # Parameter Name. Generally required at simulation-time, but `Optional` for during construction stages.
    name: Optional[str] = None


@datatype
class LinearSweep:
    """Linear Sweep"""

    start: Scalar
    stop: Scalar
    step: Scalar


@datatype
class LogSweep:
    """Logarithmic / Decade Sweep"""

    start: Scalar
    stop: Scalar
    npts: int


@datatype
class PointSweep:
    """List of Points Sweep"""

    points: List[Scalar]


# Sweep type-union
Sweep = Union[LinearSweep, LogSweep, PointSweep]


def is_sweep(val: Any) -> bool:
    return isinstance(val, Sweep.__args__)


@simattr
@datatype
class Op:
    """Operating Point Analysis"""

    name: Optional[str] = None  # Optional analysis name

    @property
    def tp(self) -> AnalysisType:
        return AnalysisType.OP


@simattr
@datatype
class Dc:
    """DC Steady-State Analysis"""

    var: Union[str, Param]  # Swept parameter, or its name
    sweep: Sweep  # Sweep values
    name: Optional[str] = None  # Optional analysis name

    @property
    def tp(self) -> AnalysisType:
        return AnalysisType.DC


@simattr
@datatype
class Ac:
    """AC Small-Signal Analysis"""

    sweep: LogSweep  # Frequency sweep values. Always log-valued.
    name: Optional[str] = None  # Optional analysis name

    @property
    def tp(self) -> AnalysisType:
        return AnalysisType.AC


@simattr
@datatype
class Tran:
    """Transient Analysis"""

    tstop: Scalar  # Stop time
    tstep: Optional[Scalar] = None  # Optional time-step recommendation
    name: Optional[str] = None  # Optional analysis name

    @property
    def tp(self) -> AnalysisType:
        return AnalysisType.TRAN


@simattr
@datatype(config=AllowArbConfig)
class Noise:
    """Noise Analysis"""

    # Output, either as a signal or pair thereof
    # NOTE: the type of `output` is really `Union[Connectable, Tuple[Connectable, Connectable]]`
    # we can't quite check this statically, but do check when serializing.
    output: Any

    # Input voltage or current source
    input_source: Union[Instance, str]

    sweep: LogSweep  # Frequency sweep values. Always log-valued.
    name: Optional[str] = None  # Optional analysis name

    @property
    def tp(self) -> AnalysisType:
        return AnalysisType.NOISE


@simattr
@datatype(config=AllowArbConfig)
class SweepAnalysis:
    """Sweep over `inner` analyses"""

    inner: List["Analysis"]  # Inner Analyses
    var: Union[str, Param]  # Sweep variable, or its name
    sweep: Sweep  # Sweep Values
    name: Optional[str] = None  # Optional analysis name

    @property
    def tp(self) -> AnalysisType:
        return AnalysisType.SWEEP


@simattr
@datatype(config=AllowArbConfig)
class MonteCarlo:
    """Add monte-carlo variations to one or more `inner` analyses."""

    inner: List["Analysis"]  # Inner Analyses
    npts: int  # Number of points
    name: Optional[str] = None  # Optional analysis name

    @property
    def tp(self) -> AnalysisType:
        return AnalysisType.MONTE


@simattr
@datatype
class CustomAnalysis:
    """String-defined, non-first-class analysis statement
    Primarily for simulator-specific specialty analyses."""

    cmd: str  # Analysis command string
    name: Optional[str] = None  # Optional analysis name

    @property
    def tp(self) -> AnalysisType:
        return AnalysisType.CUSTOM


# Analysis type-union
Analysis = Union[Op, Dc, Ac, Tran, Noise, SweepAnalysis, MonteCarlo, CustomAnalysis]


def is_analysis(val: Any) -> bool:
    return isinstance(val, Analysis.__args__)


class SaveMode(Enum):
    """Enumerated data-saving modes"""

    NONE = "none"
    ALL = "all"
    SELECTED = "selected"


# Union of "save-able" types
SaveTarget = Union[
    SaveMode,  # A `SaveMode`, e.g. `SaveMode.ALL`
    Signal,  # A single `Signal`
    List[Signal],  # A list of `Signal`s
    str,  # A signal signal-name
    List[str],  # A list of signal-names
]


@simattr
@datatype
class Save:
    """Save Control-Element
    Adds content to the target simulation output"""

    targ: SaveTarget


@simattr
@datatype(config=AllowArbConfig)
class Meas:
    """Measurement"""

    analysis: Union["Analysis", str]  # Target `Analysis`, or its type-name
    expr: str  # Measured expression. FIXME: just a string to be evaluated, no in-Python semantics
    name: Optional[str] = None  # Measurement name


@simattr
@datatype
class Include:
    """Include a File Path"""

    path: Path  # Path to include
    name: Optional[str] = None  # Name, used in class-based `Sim` definitions


@simattr
@datatype
class Lib:
    """Include a Library Section"""

    path: Path  # Path to include
    section: str  # Library Section
    name: Optional[str] = None  # Name, used in class-based `Sim` definitions


# Spice-Sim Attribute-Union
Control = Union[Include, Lib, Save, Meas, Param, Literal]


def is_control(val: Any) -> bool:
    return isinstance(val, Control.__args__)


# Define all available option types below
OptionTypes = Union[
    bool,
    Scalar,
    str,
    Literal,
]


@simattr
@datatype
class Options:
    """Simulation Options"""

    value: OptionTypes
    name: str


# Spice-Sim Attribute-Union
SimAttr = Union[Analysis, Control, Options]


def is_simattr(val: Any) -> bool:
    return isinstance(val, SimAttr.__args__)


@datatype(config=AllowArbConfig)
class Sim:
    """
    # Simulation Input

    Comprises a `Instantiable`, typically `Module`, testbench and set of simulation control input.
    Testbenches must adhere to the testbench IO interface:
    A single, width-one port, nominally named "VSS", and expected to be connected from "simulator ground" to the DUT's ground.
    """

    # Testbench, with its dependencies via `Instance`s
    tb: Instantiable
    # Simulation Control Attributes
    attrs: List[SimAttr] = field(default_factory=list)
    # Optional simulation name
    name: Optional[str] = None

    def add(self, *attrs: List[SimAttr]) -> OneOrMore[SimAttr]:
        """Add one or more `SimAttr`s to the simulation.
        Returns the inserted attributes, either as a list or a single `SimAttr`."""
        for attr in attrs:
            if not is_simattr(attr):
                raise TypeError(f"Invalid Sim attribute: {attr} being added to {self}")
            self.attrs.append(attr)
        if len(attrs) == 1:
            return attrs[0]
        return list(attrs)

    def run(self, opts: Optional[SimOptions] = None) -> SimResultUnion:
        """Invoke simulation via `vlsirtools.spice`."""
        return run(self, opts=opts)

    @property
    def Tb(self) -> "Module":
        """Get our testbench Module, with a class-style accessor name."""
        # Many testbenches "look like" classes, so we provide a class-name-style accessor
        return self.tb


def run(
    inp: OneOrMore[Sim], opts: Optional[SimOptions] = None
) -> OneOrMore[SimResultUnion]:
    """Invoke one or more `Sim`s via `vlsirtools.spice`."""

    from .proto import to_proto

    return _vlsirtools_sim(inp=to_proto(inp), opts=opts)


def _add_attr_func(name: str, cls: type):
    # Create the internal "construct + add" closure
    def _method(self, *args, **kwargs):
        inst = cls(*args, **kwargs)
        self.add(inst)
        return inst

    # Set its name to the `simattr` name
    _method.__name__ = name
    # And set it as a method on `Sim`
    setattr(Sim, name, _method)


# Add all the `simattrs` as methods on `Sim`
for name, cls in _simattrs.items():
    _add_attr_func(name, cls)


def sim(cls: type) -> Sim:
    """
    # `Sim` Definition Decorator

    Converts a class-body full of simulation attributes (`SimAttr`s) to a `Sim`.
    Example Usage:

    ```python
    import hdl21 as h
    from hdl21.sim import *

    @sim
    class MySim:
        tb = tb(name="mytb")

        x = Param(5)
        y = Param(6)
        mydc = Dc(var=x, sweep=PointSweep([1]))
        myac = Ac(sweep=LogSweep(1e1, 1e10, 10))
        mytran = Tran(tstop=11 * h.prefix.PICO)
        mysweep = SweepAnalysis(
            inner=[mytran],
            var=x,
            sweep=LinearSweep(0, 1, 2),
        )
        mymc = MonteCarlo(
            inner=[Dc(var="y", sweep=PointSweep([1]), name="swpdc")],
            npts=11,
        )
        delay = Meas(analysis=mytran, expr="trig_targ_something")
        opts = Options(reltol=1e-9)

        # Attributes whose names don't really matter can be called anything,
        # but must be *assigned* into the class, not just constructed.
        save_all = Save(SaveMode.ALL)

        # Non-`SimAttr`s such as `a_path` below will be dropped from the `Sim` definition,
        # but can be referred to by the following attributes.
        a_path = "/home/models"
        include_that_path = Include(a_path)
        fast_lib = Lib(path=a_path, section="fast")
    ```

    Class-based `Sim` definitions retain all class members which are `SimAttr`s and drop all others.
    Non-`SimAttr`-valued fields can nonetheless be handy for defining intermediate values upon which the ultimate SimAttrs depend,
    such as the `a_path` field in the example above.

    Classes decoratated by `sim` a single special required field,
    named either `tb` or `Tb`, which sets the simulation testbench.
    Valid testbenches must adhere to "the testbench IO interface":
    a single, width-one port, nominally named "VSS", and expected to be connected from "simulator ground" to the DUT's ground.
    Testbench attributes which fail to adhere to this interface will cause a `RuntimeError`.

    Several other names are disallowed in `sim` class-definitions,
    generally corresponding to the names of the `Sim` class's fields and methods.
    Disallowed names include: `["attrs", "add", "run", "namespace"]`.
    """

    if cls.__bases__ != (object,):
        raise RuntimeError(f"Invalid @hdl21.sim inheriting from {cls.__bases__}")

    protected_names = ["attrs", "add", "run", "namespace"]

    # Initialize the content of the eventual `Sim`.
    # Note we largely can't create it now because the `tb` field is required at construction time.
    name: Optional[str] = None
    tb: Optional[Instantiable] = None
    attrs: List[SimAttr] = list()

    # Any class-body content that isn't either (a) one of those special names, or (b) a `SimAttr`,
    # will be "forgotten" from the `Sim` definition.
    # This can nonetheless be handy for defining intermediate values upon which the ultimate SimAttrs depend.
    forgetme: List[Any] = list()

    # Take a lap through the class dictionary, type-check everything and assign relevant attributes to the sim
    for key, val in cls.__dict__.items():
        if key in protected_names:
            raise RuntimeError(f"Invalid field name {key} in Sim {cls}")
        elif key in ("tb", "Tb"):  # Set the test-bench attribute
            tb = val
        elif key == "name":  # Set the sim-name attribute
            name = val
        elif is_simattr(val):
            # Add to the sim-attributes list
            # Special case Python's conventional "ignored" name, the underscore.
            # Leave attributes named "_"'s `name` field set to `None`.
            if key != "_":
                val.name = key
            attrs.append(val)
        else:  # Add to the forget-list
            forgetme.append(val)

    # Validate the testbench module
    if tb is None:
        raise RuntimeError(f"No `tb` defined in Sim {cls}")
    if not is_tb(tb):
        raise RuntimeError(f"Invalid testbench {tb} in Sim {cls}")

    # Create the `Sim` object
    name = name or cls.__name__
    sim = Sim(name=name, tb=tb, attrs=attrs)

    # And return the `Sim`
    return sim



"""
# Elaboration 

Defines the primary `elaborate` method used to flesh out an in-memory `Module` or `Generator`. 
Internally defines and uses a number of hierarchical visitor-classes which traverse the hardware hierarchy, 
performing one or more transformation-passes.  
"""

# Std-Lib Imports
from typing import List, Type, TypeVar

# Local imports
from ..datatype import datatype
from .elaboratable import Elaboratable, Elaboratables, is_elaboratable

# Import all the built-in passes, and their abstract base class
from .passes import (
    ElabPass,
    InstBundleElabPass,
    Orphanage,
    ConnTypes,
    BundleFlattener,
    ResolvePortRefs,
    ArrayFlattener,
    SliceResolver,
    MarkModules,
)


ElaboratableType = TypeVar("ElaboratableType", bound=Elaboratables)


@datatype
class Elaborator:
    """
    # Elaborator

    An ordered list of `ElabPass`es.
    Each pass is implemented as a class which inherits from `ElabPass`.
    """

    passes: List[Type[ElabPass]]

    @staticmethod
    def default() -> "Elaborator":
        """Create and return the default Elaborator"""
        return Elaborator(
            passes=[
                #
                Orphanage,
                InstBundleElabPass,
                ResolvePortRefs,
                ConnTypes,
                BundleFlattener,
                ArrayFlattener,
                SliceResolver,
                #
                # A couple repeats
                #
                ConnTypes,
                Orphanage,
                #
                # And final module-marking
                #
                MarkModules,
            ]
        )

    def elaborate(self, top: ElaboratableType) -> ElaboratableType:
        """# Primary elaboration entry point
        Execute each of our `ElabPass`es in sequence."""

        # Check whether we are elaborating a single object or a list thereof
        tops: List[Elaboratable] = top if isinstance(top, List) else [top]

        # Pass `tops` through each of our passes, in order
        for elabpass in self.passes:
            tops = elabpass.elaborate(tops=tops)

        # Extract the single-element case
        if not isinstance(top, List):
            return tops[0]
        return tops


# Set the module-scope elaborator
the_global_elaborator = Elaborator.default()


def elaborate(top: ElaboratableType) -> ElaboratableType:
    """
    # Hdl21 Elaboration

    In-memory elaborates of `Module`s, calls to `Generator`s, and lists thereof.

    Optional `passes` lists the ordered `ElabPass`es to run. By default it runs the order specified by `ElabPass.default`.
    Note the order of passes is important; many depend upon others to have completed before they can successfully run.
    """
    return the_global_elaborator.elaborate(top)


def set_elaborator(e: Elaborator):
    """# Set the global `Elaborator`."""
    if not isinstance(e, Elaborator):
        raise TypeError(f"Invalid ElabPass {e}")
    global the_global_elaborator
    the_global_elaborator = e


def reset_elaborator():
    """# Reset the global `Elaborator` to its default."""
    global the_global_elaborator
    the_global_elaborator = Elaborator.default()


__all__ = [
    "elaborate",
    "Elaborator",
    "set_elaborator",
    "reset_elaborator",
    "Elaboratable",
    "Elaboratables",
    "is_elaboratable",
]



"""
# Connection Type Checking 
"""

# Std-Lib Imports
import copy
from typing import Any, Union, Dict

# PyPi Imports
from pydantic.dataclasses import dataclass

# Local imports
from ...connect import is_connectable, Connectable
from ...portref import PortRef
from ...module import Module
from ...instance import Instance
from ...noconn import NoConn
from ...bundle import (
    AnonymousBundle,
    BundleInstance,
    BundleRef,
    Bundle,
)
from ...instantiable import (
    io,
    Instantiable,
    ExternalModuleCall,
    PrimitiveCall,
)
from ..helpers.width import width, HasWidth

# Import the base class
from .base import ElabPass


"""
Connection Statuses
"""


class Valid:
    # Valid connection
    ...  # empty


@dataclass
class NoPort:
    name: str

    def __repr__(self) -> str:
        return f"Connection to non-existent Port `{self.name}`"


@dataclass
class Unconnected:
    name: str

    def __repr__(self) -> str:
        return f"Missing connection to Port `{self.name}`"


@dataclass
class InvalidType:
    msg: str  # Free-form error message

    def __repr__(self) -> str:
        return self.msg


# The union of these things is a `ConnStatus`
ConnStatus = Union[Valid, NoPort, Unconnected, InvalidType]


class ConnTypes(ElabPass):
    """
    Check for connection-type-validity on each Instance connection.

    "Connection type validity" includes:
    * All connections must be Signal-like or Bundle-like
    * Signal-valued ports must be connected to Signals of the same width
    * Bundle-valued ports must be connected to Bundles of compatible type

    Note this stage *does not* perform port-direction checking or modification.
    """

    def elaborate_module(self, module: Module) -> Module:
        """Check each Instance's connections in `module`"""

        # Check each Instance
        for inst in module.instances.values():
            self.check_instance(module, inst)

        # FIXME: whether to also check arrays.
        # This would require some more smarts about valid broadcast widths.
        # Currently they are checked by a late-stage pass after flattening.
        # for inst in module.instarrays.values():
        #     self.check_instance(module, inst)

        # No errors means it checked out, return the Module unchanged
        return module

    def check_instance(self, module: Module, inst: Instance) -> None:
        """Check the connections of `inst` in parent `module`"""
        self.stack.append(inst)

        # Get copies of both the instance's ports and connections.
        # These will be two {str: Connectable} dictionaries, who should have the same keys,
        # and each paired value should be connection-compatible.
        conns = copy.copy(inst.conns)
        io = io_for_checking(parent=module, i=inst.of)

        # Track the status of each connection, so we can report the Instance-wide state if there are errors.
        statuses: Dict[str, ConnStatus] = dict()

        for portname, port in io.items():
            # Get the corresponding connection
            conn = conns.pop(portname, None)
            if conn is None:
                statuses[portname] = Unconnected(portname)
            else:  # Check its connection-compatibility
                statuses[portname] = self.check_compatible(port, conn)

        # Now check if anything remains in `conns`, i.e. that there are invalid connections to nonexistent ports.
        for conn_name in conns.keys():
            statuses[conn_name] = NoPort(conn_name)

        # Look for any non-`Valid` connections. And if we have any, fail.
        bad_conns = {
            name: s for name, s in statuses.items() if not isinstance(s, Valid)
        }
        if bad_conns:
            msg = f"Invalid connections `{bad_conns}` on Instance `{inst.name}` in Module `{module.name}`"
            self.fail(msg)

        self.stack.pop()  # Checks out, we good, pop this Instance from our elab-stack.

    def check_bundles_compatible(
        self, bundle: Bundle, other: Union[BundleInstance, AnonymousBundle]
    ) -> ConnStatus:
        """Assert that `bundle` is compatible for connection with `other`.
        Raises a `RuntimeError` if not compatible.
        This includes key-matching and Signal-width matching, but *does not* examine Signal directions.
        """

        if isinstance(other, BundleInstance):
            other = other.of
            if other is bundle:
                return Valid()  # Same types, we good

            # Different types. Check each field in `signals` and `bundles` match.
            if sorted(bundle.signals.keys()) != sorted(other.signals.keys()):
                msg = f"Signal names do not match: {bundle.signals.keys()} != {other.signals.keys()}"
                return InvalidType(msg)

            if sorted(bundle.bundles.keys()) != sorted(other.bundles.keys()):
                msg = f"Bundle names do not match: {bundle.signals.keys()} != {other.signals.keys()}"
                return InvalidType(msg)

            # Check that each Signal is compatible
            for key, val in bundle.signals.items():
                sts = self.check_signals_compatible(val, other.signals[key])
                if not isinstance(sts, Valid):
                    return sts

            # Recursively check that each Bundle is compatible
            for key, val in bundle.bundles.items():
                sts = self.check_bundles_compatible(val.of, other.bundles[key].of)
                if not isinstance(sts, Valid):
                    return sts

            return Valid()  # Checks out, we good

        if isinstance(other, AnonymousBundle):
            # FIXME: checks on port-refs, signal-widths, etc.
            # For now this just returns success; later checks may often fail where this (eventually) should.
            return Valid()

        msg = f"Invalid connection-compatibility check between {bundle} and {other}"
        self.fail(msg)

    def check_signals_compatible(self, sig: HasWidth, other: Any) -> ConnStatus:
        """Assert that `HasWidth`s (generally `Signal`s) a and b are compatible for connection."""

        if not isinstance(other, HasWidth.__args__):
            msg = f"Invalid connection to non-Signal {other}"
            return InvalidType(msg)

        if self.get_width(sig) != self.get_width(other):
            signame = getattr(sig, "name", f"Anonymous(width={sig.width})")
            othername = getattr(other, "name", f"Anonymous(width={other.width})")
            msg = f"Signals `{signame}` and `{othername}` width mismatch: {sig.width} != {other.width}"
            return InvalidType(msg)

        return Valid()  # Checks out.

    def check_compatible(self, port: Connectable, conn: Connectable) -> ConnStatus:
        """Assert that `port` and `conn` are compatible for connection."""

        if not is_connectable(conn):
            return InvalidType(f"Invalid Connection {conn}")

        if isinstance(conn, BundleRef):
            # Recursively call this function on the ref's resolved value
            from ..helpers.resolve_ref_types import resolve_bundleref_type

            referent = resolve_bundleref_type(conn, self.fail)
            return self.check_compatible(port, referent)

        if isinstance(port, HasWidth.__args__):
            return self.check_signals_compatible(port, conn)

        if isinstance(port, BundleInstance):
            return self.check_bundles_compatible(port.of, conn)

        self.fail(f"Invalid Port {port}")

    def get_width(self, conn: Connectable) -> int:
        """Get the `width` of a conn. Fails for types which this pass is not designed to handle."""

        if isinstance(conn, (NoConn, PortRef)):
            msg = f"Internal error: {type(conn).__name__} remaining in connection-types check"
            return self.fail(msg)

        return width(conn, failer=self.fail)


def io_for_checking(parent: Module, i: Instantiable) -> Dict[str, "Connectable"]:
    """Get the relevant IOs of Instantiable `i` for checking.
    Depending on the elaboration state of `parent` and `i`, this may include the "bundled" or "flattened" IOs.
    """

    if isinstance(i, (ExternalModuleCall, PrimitiveCall)):
        # These do not have Bundle-valued ports
        return copy.copy(i.ports)

    if not isinstance(i, Module):
        raise TypeError(f"Invalid Instantiable: {i}")

    # OK we've got a Module.
    # Whether to check the "bundled" or "flattened" IO is dependent on
    # whether the *parent* and `i` have or haven't been flattened.
    parent_flattened = parent._pre_flattening_io is not None
    child_flattened = i._pre_flattening_io is not None

    if parent_flattened and not child_flattened:
        msg = f"Error child {i} has not been elaborated before parent {parent}"
        raise RuntimeError(msg)

    if parent_flattened != child_flattened:
        # Mismatch between parent and child status.
        # Return the child IOs from just *before* the flattening pass.
        return copy.copy(i._pre_flattening_io)

    # Parent and child statuses match, whether flattened or not.
    # Return the child IOs as they are now.
    return io(i)



""" 
# Delay Measurements and Sims 

The many, many-th iteration of taking a `Module` comprised of digital logic, 
setting the states of all but one input, stepping the other, and measuring the delay(s) 
to each of its outputs. 
"""

from copy import copy
from enum import Enum
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple

import vlsir.spice_pb2 as vsp
from vlsirtools.spice import sim_data as sd, SimResultUnion

# Local Imports
from ..module import Module
from ..signal import Signal, Port, Visibility, PortDir, _copy_to_internal
from ..instantiable import Instantiable
from ..prefix import Prefix
from ..primitives import Vdc, Vpu, Cap
from .data import ParamVal, Sim, SimAttr


class LogicState(Enum):
    LOW = False
    HIGH = True


class Transition(Enum):
    RISING = "rising"
    FALLING = "falling"


@dataclass
class DelaySimParams:
    dut: Instantiable  # The design under test

    primary_input: Signal  # The primary input to be tested
    other_inputs: Dict[str, LogicState]  # Logical values of other inputs

    input_trans: Transition  # Primary-input transition direction

    vlo: ParamVal  # Logic-low level
    vhi: ParamVal  # Logic-high level

    # Mapping from supply (names) to values
    supplies: Dict[str, ParamVal] = field(default_factory=dict)
    # List of ground signals (names)
    grounds: List[Signal] = field(default_factory=list)

    # Capacitance on each output, keyed by name
    load_caps: Dict[str, ParamVal] = field(default_factory=dict)
    # Default capacitance on each output, if not specified in `load_caps`
    default_load_cap: ParamVal = 1 * Prefix.ATTO

    trf: ParamVal = 1 * Prefix.PICO  # Input rise / fall time
    tstop: ParamVal = 1 * Prefix.NANO  # Sim stop time
    tstep: Optional[ParamVal] = 1 * Prefix.PICO  # Recommended sim timestep

    # Sadly we don't really know how to pass these into `Sim` yet, without some help here.
    pathsep: str = "."  # Hierarchical path separator
    # Expression delimiters, opening and closing. Pair of single quotes by default.
    expr_delim: Tuple[str, str] = ("'", "'")

    # Additional simulation attributes (options, saves, etc)
    attrs: List[SimAttr] = field(default_factory=list)


@dataclass
class DelaySimResult:
    """Results from a `DelaySim`"""

    # Per-output delays, keyed by output signal name
    delays: Dict[str, Optional[float]]


def delays(p: DelaySimParams, opts: Optional[vsp.SimOptions] = None) -> DelaySimResult:
    """Create and run a delay `Sim`.

    Executes four discrete steps, all defined in this module:

    * `create_sim`
    * `run` the generated `Sim`
    * `get_meas` is a short step to gather the measurement dictionary. It will (should) eventually disappear.
    * `collect_result`

    Each individually for cases needing to "get in between" them.
    """
    sim = create_sim(p)
    results = sim.run(opts=opts)
    meas = get_meas(results)
    return collect_result(p, meas)


def create_sim(p: DelaySimParams) -> Sim:
    """Create a delay `Sim`"""

    # Create the testbench `Module` and `Sim` input
    tb = Module(name=f"{p.dut.name}Tb")
    tb.vss = Port(width=1)
    sim = Sim(tb=tb, attrs=p.attrs)
    # Add the primary transient analysis
    tran = sim.tran(
        tstop=p.tstop, tstep=p.tstep, name=f"tran_delay_{p.primary_input.name}"
    )

    # Instantiate the DUT
    tb.dut = p.dut()

    # Create the primary input Signal
    sig = copy(p.primary_input)
    if sig.width != 1:
        raise RuntimeError(f"Unsupported `Delay` sim for bus input {inp}")
    sig.vis = Visibility.INTERNAL
    sig.direction = PortDir.NONE
    tb.add(sig)
    tb.dut.connect(sig.name, sig)

    # Create its pulse-source
    (v1, v2) = (p.vlo, p.vhi) if p.input_trans == Transition.RISING else (p.vhi, p.vlo)
    vpulse_params = Vpu.Params(
        v1=v1,
        v2=v2,
        delay=0 * Prefix.UNIT,
        period=2 * Prefix.UNIT,
        rise=p.trf,
        fall=p.trf,
        width=1 * Prefix.UNIT,
    )
    vpulse = Vpu(vpulse_params)(p=sig, n=tb.vss)
    tb.add(vpulse, name=f"v{p.primary_input.name}")

    # Add a measurement of the inital primary-input crossing time
    cross = "rise" if p.input_trans == Transition.RISING else "fall"
    sim.meas(
        analysis=tran,
        name=f"tcross_primary_input",
        expr=f"when V(xtop{p.pathsep}{p.primary_input.name})={p.expr_delim[0]}vdd/2{p.expr_delim[1]} {cross}=1",  # FIXME: simulator-specific
    )

    for gndsig in p.grounds:
        # Make each ground connection to test-bench VSS
        tb.dut.connect(gndsig.name, tb.vss)

    for vname, vval in p.supplies.items():
        # Create and connect each supply voltage
        # And add a voltage-source instance to drive it
        sig = p.dut.get(vname)
        if sig is None:
            raise RuntimeError(f"Invalid Supply {vname} in creating delay for {p.dut}")
        sig = tb.add(_copy_to_internal(sig))
        tb.add(Vdc(Vdc.Params(dc=vval))(p=sig, n=tb.vss), name=f"v{sig.name}")
        tb.dut.connect(sig.name, sig)

    dut_inputs = [s for s in p.dut.ports.values() if s.direction == PortDir.INPUT]
    for inp in dut_inputs:
        if inp == p.primary_input:
            continue  # Skip the primary input

        # Create a Signal for each DUT input
        inp = tb.add(_copy_to_internal(inp))
        if inp.width != 1:
            raise RuntimeError(f"Unsupported `Delay` sim for bus input {inp}")
        tb.dut.connect(inp.name, inp)

        # Get its logic state
        state = p.other_inputs.get(inp.name, None)
        if state is None:
            raise ValueError(f"No LogicState defined for input {inp}")
        if state == LogicState.HIGH:
            vdc = p.vhi
        elif state == LogicState.LOW:
            vdc = p.vlo
        else:
            raise ValueError(f"Invalid LogicState {state} for Input {inp}")

        # And add a voltage-source instance to drive it
        tb.add(Vdc(Vdc.Params(dc=vdc))(p=inp, n=tb.vss), name=f"v{inp.name}")

    # Create everything associated with each output - load caps and measurements
    dut_outputs = [s for s in p.dut.ports.values() if s.direction == PortDir.OUTPUT]
    for out in dut_outputs:
        out = tb.add(_copy_to_internal(out))
        if out.width != 1:
            raise RuntimeError(f"Unsupported `Delay` sim for bus output {inp}")
        tb.dut.connect(out.name, out)

        # Sort out the load capacitance value, from either the signal-specific dict, or the default value
        c = p.load_caps.get(out.name, p.default_load_cap)

        # Add a load cap instance
        tb.add(Cap(Cap.Params(c=c))(p=out, n=tb.vss), name=f"c{out.name}")

        # Add a delay measurement
        sim.meas(
            analysis=tran,
            name=f"tcross_{out.name}",
            expr=f"when V(xtop{p.pathsep}{out.name})={p.expr_delim[0]}vdd/2{p.expr_delim[1]} cross=1",  # FIXME: simulator-specific
        )
        sim.meas(
            analysis=tran,
            name=f"tdelay_{out.name}",
            expr=f"PARAM={p.expr_delim[0]}tcross_{out.name}-tcross_primary_input{p.expr_delim[1]}",  # FIXME: simulator-specific
        )

    # Aaaaaaand return the sim already!
    return sim


def collect_result(
    p: DelaySimParams, meas: Dict[str, Optional[float]]
) -> DelaySimResult:
    """
    Collect measured output delays
    After simulation has run and data has been pulled back into memory,
    extract the delay value for each of `p.module`'s outputs.
    """
    delays = {}
    dut_outputs = [s for s in p.dut.ports.values() if s.direction == PortDir.OUTPUT]
    for out in dut_outputs:
        out_meas = meas.get(f"tdelay_{out.name}", None)
        if out_meas is None:
            delays[out.name] = None
        else:
            try:
                delays[out.name] = float(out_meas)
            except:
                delays[out.name] = None
    return DelaySimResult(delays)


def get_meas(results: SimResultUnion) -> Dict[str, Optional[float]]:
    """Get the `measurements` dict from either of the `VlsirTools` sim-results types.
    FIXME: this should get pushed down to `VlsirTools`."""

    if isinstance(results, sd.SimResult):
        return results.an[0].measurements
    if isinstance(results, vsp.SimResult):
        return results.an[0].tran.measurements
    raise TypeError



"""
Flatten a module by moving all nested instances, ports and signals to the root level.

See function `flatten()` for details.
"""

import copy
from pydantic.dataclasses import dataclass
from dataclasses import field, replace
from typing import Dict, Generator, List, Optional

import hdl21 as h
from .signal import _copy_to_internal
from .datatype import AllowArbConfig


def _walk_conns(conns, parents=tuple()):
    for key, val in conns.items():
        if isinstance(val, dict):
            yield from _walk_conns(val, parents + (key,))
        else:
            yield parents + (key,), val


def _flat_conns(conns):
    return {":".join(reversed(path)): value.name for path, value in _walk_conns(conns)}


@dataclass(config=AllowArbConfig)
class FlattenedInstance:
    """
    # Flattened Instance
    An instance that survives flattening, because its contents are either (a) Primitive or (b) External.
    """

    inst: h.Instance
    path: List[h.Instance] = field(default_factory=list)
    conns: Dict[str, h.Signal] = field(default_factory=dict)

    def __post_init__(self):
        # Assert that this instance's target is either a primitive, or external
        if not isinstance(self.inst.of, (h.PrimitiveCall, h.ExternalModuleCall)):
            raise ValueError(f"Invalid flattened instance {self}")

    def __str__(self):
        name = self.make_name()
        path = [p.name for p in self.path]
        conns = _flat_conns(self.conns)
        return str(dict(name=name, path=path, conns=conns))

    def make_name(self):
        return ":".join([p.name or "_" for p in self.path])


def walk(
    m: h.Module,
    parents: List[h.Instance],
    conns: Optional[Dict[str, h.Signal]] = None,
) -> Generator[FlattenedInstance, None, None]:
    if conns is None:
        conns = {**m.signals, **m.ports}
    for inst in m.instances.values():
        new_conns = {}
        new_parents = parents + [inst]
        for src_port_name, sig in inst.conns.items():
            if isinstance(sig, h.Signal):
                key = sig.name
            elif isinstance(sig, (h.Slice, h.Concat)):
                msg = f"Flattening `Slice` and `Concat` is not (yet) supported"
                raise NotImplementedError(msg)
            elif isinstance(sig, (h.PortRef, h.BundleInstance, h.AnonymousBundle)):
                # This shouldn't happen in normal use, but could in principle if
                # someone e.g. calls this `walk` function directly.
                msg = f"Error: {sig} should not have reached this stage in flattening"
                raise RuntimeError(msg)
            else:
                raise TypeError(f"Invalid connection {sig}")

            new_sig_name = ":".join([p.name for p in parents] + [key])
            if key in conns:
                target_sig = conns[key]
            elif key in m.signals:
                target_sig = replace(
                    _copy_to_internal(m.signals[key]), name=new_sig_name
                )
            elif key in m.ports:
                target_sig = replace(_copy_to_internal(m.ports[key]), name=new_sig_name)
            else:
                raise ValueError(f"signal {key} not found")
            new_conns[src_port_name] = target_sig

        if isinstance(inst.of, h.PrimitiveCall):
            yield FlattenedInstance(inst, new_parents, new_conns)
        else:
            yield from walk(inst.of, new_parents, new_conns)


def _find_signal_or_port(m: h.Module, name: str) -> h.Signal:
    """Find a signal or port by name"""

    port = m.ports.get(name, None)
    if port is not None:
        return port

    sig = m.signals.get(name, None)
    if sig is not None:
        return sig

    raise ValueError(f"Signal {name} not found in module {m.name}")


def is_flat(m: h.Instantiable) -> bool:
    """Boolean indication of whether `Instantiable` `m` is already flat."""

    if isinstance(m, (h.PrimitiveCall, h.ExternalModuleCall)):
        return True
    if isinstance(m, h.Module):
        instancelike = (
            list(m.instances.values())
            + list(m.instarrays.values())
            + list(m.instbundles.values())
        )
        return all(
            isinstance(inst.of, (h.PrimitiveCall, h.ExternalModuleCall))
            for inst in instancelike
        )
    raise TypeError(f"Invalid `Instantiable` argument to `is_flat`: {m}")


def flatten(m: h.Instantiable) -> h.Instantiable:
    r"""Flatten a module by moving all nested instances, ports and signals to the root level.

    For example, if we have a buffer module with two inverters, `inv_1` and `inv_2`, each with
    two transistors, `pmos` and `nmos`, the module hierarchy looks like this:
    All signals and ports will be moved to the root level too.

    ```
    buffer
    ├── inv_1
    │   ├── pmos
    │   └── nmos
    └── inv_2
        ├── pmos
        └── nmos
    ```

    This function will flatten the module to the following structure:

    ```
    buffer_flat
    ├── inv_1:pmos
    ├── inv_1:nmos
    ├── inv_2:pmos
    └── inv_2:nmos
    ```

    Nested signals and ports will be renamed and moved to the root level as well. For example, say a
    module with two buffers, `buffer_1` and `buffer_2`, each with two inverters. On the top level,
    the original ports `vdd`, `vss`, `vin` and `vout` are preserved. Nested signals and ports such
    the connection between two buffers, the internal signal between two inverters in buffer 1 will
    be renamed and moved to the root level as `buffer_1_vout`, `buffer_1:inv_1_vout`,
    `buffer_2:inv_1_vout`.

    See tests/test_flatten.py for more examples.
    """

    m = h.elaborate(m)
    if is_flat(m):
        return m

    # recursively walk the module and collect all primitive instances
    nodes: List[FlattenedInstance] = list(walk(m, parents=[]))

    # Create our new, flattened Module
    # Note that by virtue of going through elaboration above, `m.name` should be set.
    # Check for it nonetheless, and raise an error if not.
    if m.name is None:
        raise ValueError(f"Anonymous Module {m} cannot be flattened. (Give it a name.)")
    new_module = h.Module(m.name + "_flat")
    for port in m.ports.values():
        new_module.add(copy.copy(port))

    # add all signals to the root level
    for n in nodes:
        for sig in n.conns.values():
            sig_name = sig.name
            if sig_name not in new_module.ports:
                new_module.add(copy.copy(sig))

    # add all connections to the root level with names resolved
    for n in nodes:
        new_inst = new_module.add(n.inst.of(), name=n.make_name())

        for src_port_name, sig in n.conns.items():
            matching_sig = _find_signal_or_port(new_module, sig.name)
            new_inst.connect(src_port_name, matching_sig)

    return new_module



""" 
# `hdl21.Bundle` and Related Classes 

Structured connection objects, instances thereof, and associated utilities.
"""

# Std-Lib Imports
import warnings
from copy import copy
from enum import Enum, EnumMeta
from typing import Optional, Union, Any, Dict, Set, List, ClassVar

# Local Imports
from . import attrmagic
from .role import Role, RoleSet
from .connect import connectable, is_connectable
from .sliceable import sliceable
from .concat import concatable
from .signal import Signal
from .visibility import Visibility
from .props import Properties


def getattr_bundle_refs(cls: type) -> type:
    """Decorator to add the "__getattr__ generates BundleRefs" functionality to `cls`.
    Adds the `_bundle_ref` method and `__getattr__` access to it."""

    # First check and fail if any of the methods to be defined here are already defined elsewhere
    defined_here = ["__getattr__"]
    for key in defined_here:
        if key in cls.__dict__:
            msg = f"Invalid modification of {cls} with `@getattr_bundle_refs`: {key} is already defined."
            raise RuntimeError(msg)

    def __getattr__(self, key: str) -> Any:
        """BundleRef access by getattr"""
        if key.startswith("_") or not self.__getattribute__("_initialized"):
            # Bootstrapping phase: do regular getattrs to get started
            return object.__getattribute__(self, key)

        if key in self.__getattribute__("_specialcases"):  # Special case(s)
            return object.__getattribute__(self, key)

        # After elaboration, the fancy BundleRef creation below goes away. Only return ready-made refs.
        if self.__getattribute__("_elaborated"):
            refs = self.__getattribute__("refs_to_me")
            if key in refs.keys():
                return refs[key]
            raise AttributeError(f"No attribute {key} for {self}")

        # Fell through all those cases. Fancy `BundleRef` generation time!
        # Return a `BundleRef`, creating one if necessary.
        return _bundle_ref(self, key)

    cls.__getattr__ = __getattr__
    cls.__getattr_bundle_refs__ = True

    # And don't forget to return it!
    return cls


def has_getattr_bundle_refs(obj: Any) -> bool:
    """Boolean indication of "getattr bundle refs" functionality"""
    return getattr(obj, "__getattr_bundle_refs__", False)


@getattr_bundle_refs
@connectable
@attrmagic.only_set_known_attrs
@attrmagic.init
class BundleInstance:
    """# Instance of a `Bundle`
    Generally in a `Module` or another `Bundle`"""

    _specialcases = [
        "name",
        "of",
        "port",
        "role",
        "src",
        "dest",
        "desc",
        "props",
        "refs_to_me",
    ]

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        of: "Bundle",
        port: bool = False,
        flipped: bool = False,
        role: Union[Role, Enum, None] = None,
        src: Union[Role, Enum, None] = None,
        dest: Union[Role, Enum, None] = None,
        desc: Optional[str] = None,
    ):
        self.name = name
        self.of = of
        self.port = port  # FIXME: make this a `Visibility`
        self.flipped = flipped
        self.role = role
        self.src = src
        self.dest = dest
        self.desc = desc
        # References handed out to our children
        self.refs_to_me: Dict[str, "BundleRef"] = dict()
        self.props: Properties = Properties()
        # Connected port references
        self._connected_ports: Set["PortRef"] = set()
        self._parent_module: Optional["Module"] = None
        self._elaborated = False
        self._initialized = True

    """ Special Methods """

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name} of={self.of})"

    def __rmul__(self, num: int) -> List["Self"]:
        """# Right multiplication. Creates `num` copies of ourselves."""
        if not isinstance(num, int):
            return NotImplemented
        return [copy(self) for _ in range(num)]


# Type-alias for HDL objects storable as `Module` attributes
BundleAttr = Union[Signal, BundleInstance]


def is_bundle_attr(val: Any) -> bool:
    """Boolean indication of whether `val` is a valid `hdl21.Bundle` attribute."""
    return isinstance(val, BundleAttr.__args__)


def assert_bundle_attr(b: "Bundle", val: Any) -> None:
    """Raise a TypeError if `val` is not a valid attribute."""
    if not is_bundle_attr(val):
        msg = f"Invalid attribute {val} added to {b}"
        raise TypeError(msg)


_banned = ["signals", "bundles", "namespace"]


@attrmagic.init
class Bundle:
    """
    # hdl21 Bundle

    Bundles are structured hierarchical connection objects which include Signals and other Bundles.
    """

    def __init__(self, *, name: Optional[str] = None):
        self.name: Optional[str] = name
        self.roles: Optional[RoleSet] = None
        self.signals: Dict[str, "Signal"] = dict()
        self.bundles: Dict[str, "BundleInstance"] = dict()
        self.namespace: Dict[str, "BundleAttr"] = dict()  # Combination of all these
        self.props: Properties = Properties()
        self._elaborated = (
            False  # FIXME: unify whether these are boolean, or Optional objects
        )
        self._initialized = True  # Done running __init__, enable some magic

    """
    The public `Bundle` API. Just `add` and `get`. 
    To keep the space of valid names for post-fix dot access as large as possible, 
    the `Bundle` class namespace is kept *as small* as possible. 
    """

    def add(self, val: BundleAttr, *, name: Optional[str] = None) -> BundleAttr:
        """Add a Bundle attribute.

        `Bundle.add` allows for programmatic insertion of attributes whose names are not legal Python identifiers,
        such as keywords ('in', 'from') and those including invalid characters.

        The added object `val` is also provided as the return value, enabling chaining-style usages such as
        ```python
        instance.inp = MyBundle.add(h.Input(name="in", width=5))
        ```
        and similar.

        Attribute naming comes from one of either:
        * `val`'s `name` attribute, if it has one, or
        * The optional `name` argument.
        Only one of the two name-sources is allowed per call.
        """

        # Check it's a valid attribute-type
        assert_bundle_attr(self, val)

        # Now sort out naming. We get two name-sources:
        # (a) the function-argument `name` and (b) the value's `name` attribute.
        # One or the other (and not both) must be set.
        if name is None and val.name is None:  # Neither set, fail.
            msg = f"Anonymous attribute {val} cannot be added to Bundle {self.name}"
            raise RuntimeError(msg)
        if name is not None and val.name is not None:  # Both set, fail.
            msg = f"{val} with conflicting names {name} and {val.name} cannot be added to Bundle {self.name}"
            raise RuntimeError(msg)
        if name is not None:  # One or the other set - great.
            val.name = name

        # Now `val.name` is set appropriately.
        # Add it to our type-based containers, and return it.
        return _add(bundle=self, val=val)

    def get(self, name: str) -> Optional[BundleAttr]:
        """Get attribute `name`. Returns `None` if not present.
        Note unlike Python built-ins such as `getattr`, `get` returns solely
        from the HDL namespace-worth of `BundleAttr`s."""
        ns = self.__getattribute__("namespace")
        return ns.get(name, None)

    @property
    def Roles(self):
        """# Roles Accessor Property"""
        # Roles often look like a class, so they have a class-style name-accessor
        return self.roles

    """
    Special Methods
    """

    def __setattr__(self, key: str, val: Any) -> None:
        """Set-attribute over-ride, organizing into type-based containers"""

        if key.startswith("_") or not getattr(self, "_initialized", False):
            # Bootstrapping phase. Pass along to "regular" setattr.
            return super().__setattr__(key, val)

        if key in _banned:
            msg = f"Error attempting to over-write protected attribute {key} of Module {self}"
            raise RuntimeError(msg)
        # Special case(s)
        if key == "name":
            return super().__setattr__(key, val)
        if key == "roles":
            if isinstance(val, EnumMeta):
                # FIXME: deprecation warning and inline conversion here for now!
                msg = f"Bundle roles should be a `RoleSet`. Use `RoleSet.from_enum` to convert an `Enum` class!"
                warnings.warn(DeprecationWarning(msg))
                val = RoleSet.from_enum(val)
            if not isinstance(val, RoleSet):
                raise TypeError(f"Bundle roles must be an `RoleSet`, not {val}")

            return super().__setattr__(key, val)

        # Check it's a valid attribute-type
        assert_bundle_attr(self, val)

        # Checks out! Name `val` and add it to our type-based containers.
        val.name = key
        _add(bundle=self, val=val)
        return None

    def __getattr__(self, key):
        ns = self.__getattribute__("namespace")
        if key in ns:
            return ns[key]
        return object.__getattribute__(self, key)

    def __call__(self, **kwargs):
        """Calls to Bundles return Bundle Instances"""
        return BundleInstance(of=self, **kwargs)

    def __init_subclass__(cls, *_, **__):
        """Sub-Classing Disable-ization"""
        msg = f"Error attempting to create {cls.__name__}. Sub-Typing hdl21.Bundle is not supported."
        raise RuntimeError(msg)

    def __repr__(self) -> str:
        if self.name:
            return f"Bundle(name={self.name})"
        return f"Bundle(_anon_)"


def _add(bundle: Bundle, val: BundleAttr) -> BundleAttr:
    """Internal `Module.add` and `Module.__setattr__` implementation.
    Primarily sort `val` into one of our type-based containers.
    Layers above `_add` must ensure that `val` has its `name` attribute before calling this method.
    """

    if bundle._elaborated:  ## FIXME: is not None:
        raise RuntimeError(f"Cannot add {val} to {bundle} after elaboration.")

    # Sort out which of our type-based containers to add `val` to.
    if isinstance(val, Signal):
        type_ctr = bundle.signals
    elif isinstance(val, BundleInstance):
        type_ctr = bundle.bundles
    elif isinstance(val, Role):
        type_ctr = bundle.roles.inner
    else:
        # The next line *should* never be reached, as outer layers should have checked `_is_bundle_attr`.
        # Nonetheless gotta raise an error if we get here, somehow.
        msg = f"Invalid Bundle attribute {val} for {bundle}"
        raise TypeError(msg)

    # Add it to the bundle namespace, and the type-specific container
    type_ctr[val.name] = val
    bundle.namespace[val.name] = val

    # Give it a reference to us
    val._parent_bundle = bundle

    # And return our newly-added attribute
    return val


def bundle(cls: type) -> Bundle:
    """# Bundle Definition Decorator

    Converts a class-body full of Bundle-storable attributes (Signals, other Bundles) to an `hdl21.Bundle`.
    Example Usage:

    ```python
    import hdl21 as h

    @h.bundle
    class Diff:
        p = h.Signal()
        n = h.Signal()

    @h.bundle
    class DisplayPort:
        main_link = Diff()
        aux = h.Signal()
    ```

    Bundles may also define a `Roles` enumeration, inline within this class-body.
    `Roles` are optional pieces of enumerated endpoint-labels which aid in dictating `Signal` directions.
    Each `Signal` accepts optional source (`src`) and destination (`dst`) fields which (if set) must be one of these roles.

    ```python
    import hdl21 as h
    from enum import Enum, auto

    @h.bundle
    class HasRoles:
        class Roles(Enum):
            HOST = auto()
            DEVICE = auto()

        tx = h.Signal(src=Roles.HOST, dest=Roles.DEVICE)
        rx = h.Signal(src=Roles.DEVICE, dest=Roles.HOST)
    ```

    Bundle roles may be set by defining either a (lower case) `roles` or an
    (upper case) `Roles` attribute within the class body.
    In the resultant `Bundle` they will be accessible as either.
    """
    if cls.__bases__ != (object,):
        raise RuntimeError(f"Invalid @hdl21.bundle inheriting from {cls.__bases__}")

    # Create the Bundle object
    bundle = Bundle(name=cls.__name__)

    protected_names = ["signals", "bundles"]
    # Any class-body content that isn't a `ModuleAttr` will be "forgotten" from the `Module` definition.
    # This can nonetheless be handy for defining intermediate values upon which the ultimate Module attributes depend.
    forgetme: List[Any] = list()
    # Collect a list of all declared `Role`s
    roles_dict: Dict[str, Role] = dict()

    # Take a lap through the class dictionary, type-check everything and assign relevant attributes to the bundle
    for key, val in cls.__dict__.items():
        if key in protected_names:
            raise RuntimeError(f"Invalid field name {key} in bundle {cls}")
        elif key == "roles" or key == "Roles":
            # Special-case the upper-cased `Roles`, as it'll often be a class-def
            setattr(bundle, "roles", val)
        elif isinstance(val, Role):
            roles_dict[key] = val
        elif is_bundle_attr(val):
            setattr(bundle, key, val)
        else:  # Add to the forget-list
            forgetme.append(val)

    # Sort out the role definitions.
    # This can come either from a single `roles`/`Roles` attribute, or from a list of `Role`s, but not both.
    if roles_dict:
        if bundle.roles is not None:
            msg = f"Error attempting to define roles for {bundle} when it already has roles {bundle.roles}"
            raise RuntimeError(msg)
        # Defined via list. Convert to a `RoleSet`.
        bundle.roles = RoleSet.from_dict(roles_dict)

    # And return the bundle
    return bundle


@attrmagic.no_setattr
@attrmagic.init
@connectable
class AnonymousBundle:
    """# Anonymous Connection Bundle
    Commonly used for "collecting" Signals into `h.Bundle`s,
    or for re-jiggering connections between `h.Bundle`s."""

    def __init__(self, **kwargs: Dict[str, "Connectable"]):
        # Create our internal structures
        # The core namespace of Signals, other Bundles, and any other Connectables
        self._namespace: Dict[str, "Signal"] = dict()
        # Connected port references
        self._connected_ports: Set["PortRef"] = set()

        # And add each keyword-arg
        for key, val in kwargs.items():
            self.add(key, val)

        self._initialized = True

    def add(self, name: str, val: BundleAttr) -> BundleAttr:
        """Add attribute `val` to our namespace."""

        if name in self._namespace:
            raise RuntimeError(f"Duplicate attribute {name} in {self}")
        if not is_connectable(val):
            raise TypeError(f"Invalid Bundle attribute {val} for {self}")
        self._namespace[name] = val
        return val

    def get(self, name: str) -> Optional["Connectable"]:
        """Get attribute `name`. Returns `None` if not present.
        Note unlike Python built-ins such as `getattr`, `get` returns solely
        from the HDL namespace-worth of attributes."""
        return self._namespace.get(name, None)


def bundlize(**kwargs) -> AnonymousBundle:
    """# Bundle-ize
    "Collect" a group of Signals and Bundle instances into an anonymous Bundle."""
    return AnonymousBundle(**kwargs)


@getattr_bundle_refs
@concatable
@sliceable
@connectable
@attrmagic.only_set_known_attrs
@attrmagic.init
class BundleRef:
    """Reference into a Bundle Instance"""

    _specialcases: ClassVar[List[str]] = [
        "parent",
        "attrname",
        "path",
        "root",
        "refs_to_me",
        "resolved",
    ]

    def __init__(
        self,
        parent: Union["BundleInstance", "BundleRef"],  # Parent Bundle
        attrname: str,  # Attribute name
    ):
        self.parent = parent
        self.attrname = attrname

        self.resolved: Union[None, "Signal", "BundleInstance"] = None
        # References handed out to our children
        self.refs_to_me: Dict[str, "BundleRef"] = dict()
        self._width: Optional[int] = None  # FIXME: remove?
        self._slices: Set["Slice"] = set()
        self._concats: Set["Concat"] = set()
        self._connected_ports: Set["PortRef"] = set()

        self._elaborated = False
        self._initialized = True

    def __eq__(self, other) -> bool:
        """Port-reference equality requires *identity* between parents
        (and of course equality of attribute-name)."""
        return self.inst is other.inst and self.attrname == other.attrname

    def __hash__(self):
        """Hash references as the tuple of their instance-address and name"""
        return hash((id(self.inst), self.attrname))

    def path(self) -> List[str]:
        """Get the path to this potentially nested reference."""
        if isinstance(self.parent, BundleRef):
            return self.parent.path() + [self.attrname]
        if isinstance(self.parent, BundleInstance):
            return [self.attrname]
        raise TypeError(f"Invalid parent {self.parent} for {self}")

    def root(self) -> "BundleInstance":
        """Get the root `BundleInstance` of this potentially nested reference."""
        if isinstance(self.parent, BundleRef):
            return self.parent.root()
        if isinstance(self.parent, BundleInstance):
            return self.parent
        raise TypeError(f"Invalid parent {self.parent} for {self}")

    def __repr__(self):
        return f"{self.__class__.__name__}(root={self.root()} path={self.path()})"


def _bundle_ref(self: Union[BundleRef, BundleInstance], key: str) -> BundleRef:
    """Return a reference to name `key`, creating it if necessary."""

    # Check in our existing references
    bundle_refs = self.__getattribute__("refs_to_me")
    if key in bundle_refs:
        return bundle_refs[key]

    # New reference; create, add, and return it
    bundle_ref = BundleRef(parent=self, attrname=key)
    bundle_refs[key] = bundle_ref
    return bundle_ref


def flippable(b: Bundle) -> bool:
    """
    # Boolean indication of whether a Bundle is flippable.
    Requires that all Signals in the Bundle,
    and all Signals in all its sub-bundles, have port-visibility.
    """
    return all([s.vis == Visibility.PORT for s in b.signals.values()]) and all(
        [flippable(bi.of) for bi in b.bundles.values()]
    )


def flipped(bi: BundleInstance) -> BundleInstance:
    """# Create a flipped copy of a BundleInstance"""
    cp = copy(bi)
    cp.flipped = not cp.flipped
    return cp


__all__ = [
    "Bundle",
    "bundle",
    "BundleInstance",
    "AnonymousBundle",
    "bundlize",
    "flipped",
    "BundleRef",  # FIXME: remove
]



"""
# VLSIR ProtoBuf Schema Export
"""

# Std-Lib Imports
from typing import List, Union, Sequence
from decimal import Decimal

# VLSIR Import
import vlsir
import vlsir.circuit_pb2 as vckt
import vlsir.spice_pb2 as vsp

# Local Imports
from ..one_or_more import OneOrMore
from . import data
from ..literal import Literal
from ..prefix import Prefixed
from ..scalar import Scalar
from ..signal import Signal
from ..connect import is_connectable


def to_proto(inp: OneOrMore[data.Sim]) -> OneOrMore[vsp.SimInput]:
    """Convert a `Sim` to a VLSIR `SimInput`"""
    from ..proto import to_proto as module_to_proto

    # Convert the testbench module and all its dependencies into a `Package`
    single_input = False
    if not isinstance(inp, Sequence):
        inp = [inp]
        single_input = True

    # Co-export, in the process co-elaborating, all provided testbenches
    pkg: vckt.Package = module_to_proto([i.tb for i in inp])

    # Export each individual `Sim`
    results = [SimProtoExporter(sim=s, pkg=pkg).export() for s in inp]

    # And return the result in the form in which it came
    if single_input:
        return results[0]
    return results


class SimProtoExporter:
    """Simulation Protobuf Exporter"""

    def __init__(self, sim: data.Sim, pkg: vckt.Package):
        self.sim = sim  # Store the input `Sim`
        self.pkg = pkg
        self.inp = None  # Initialize our resultant `vsp.SimInput`
        self.analysis_count = 0

    def export(self) -> vsp.SimInput:
        """Primary export method. Converts `sim.tb` and its dependencies to a `Package`,
        and all other `SimAttr`s to VLSIR attributes."""

        from ..instantiable import qualname

        # Now that the testbench has been elaborated, ensure it adheres to the testbench interface.
        if not data.is_tb(self.sim.tb):
            raise RuntimeError(f"Invalid Testbench {self.sim.tb} for Simulation")

        # Create our `SimInput`, and provide it with a "link" to the testbench, via its name
        self.inp = vsp.SimInput(pkg=self.pkg, top=qualname(self.sim.tb))

        # Export each simulation attribute
        for attr in self.sim.attrs:
            self.export_attr(attr)
        return self.inp

    def export_attr(self, attr: data.SimAttr) -> None:
        """Export a Sim Attribute. Primarily dispatches across internal union-types."""
        from .data import is_analysis, is_control

        if isinstance(attr, data.Options):
            self.inp.opts.append(export_options(attr))
        elif is_analysis(attr):
            self.inp.an.append(self.export_analysis(attr))
        elif is_control(attr):
            self.inp.ctrls.append(export_control(attr))
        else:
            raise TypeError(f"Invalid SimAttr: {attr}")

    def export_analysis(self, an: data.Analysis) -> vsp.Analysis:
        """Export an `Analysis`, largely dispatching across types and re-assembling into a `vsp.Analysis`."""
        if isinstance(an, data.Op):
            return vsp.Analysis(op=self.export_op(an))
        elif isinstance(an, data.Dc):
            return vsp.Analysis(dc=self.export_dc(an))
        elif isinstance(an, data.Ac):
            return vsp.Analysis(ac=self.export_ac(an))
        elif isinstance(an, data.Tran):
            return vsp.Analysis(tran=self.export_tran(an))
        elif isinstance(an, data.Noise):
            return vsp.Analysis(noise=self.export_noise(an))
        elif isinstance(an, data.SweepAnalysis):
            return vsp.Analysis(sweep=self.export_sweep_analysis(an))
        elif isinstance(an, data.MonteCarlo):
            return vsp.Analysis(monte=self.export_monte(an))
        elif isinstance(an, data.CustomAnalysis):
            return vsp.Analysis(custom=self.export_custom_analysis(an))
        else:
            raise TypeError(f"Invalid Analysis {an}")

    def next_analysis_name(self) -> str:
        """Create a name for the next user-unnamed Analysis.
        Format: `Analysis{num}`, where `num` increases across all analyses."""
        name = f"Analysis{self.analysis_count}"
        self.analysis_count += 1
        return name

    def export_op(self, op: data.Op) -> vsp.OpInput:
        """Export an operating point analysis"""
        analysis_name = op.name or self.next_analysis_name()
        return vsp.OpInput(
            analysis_name=analysis_name,
            ctrls=[],  # FIXME: analysis-specific controls
        )

    def export_dc(self, dc: data.Dc) -> vsp.DcInput:
        """Export a DC analysis"""
        analysis_name = dc.name or self.next_analysis_name()
        return vsp.DcInput(
            analysis_name=analysis_name,
            indep_name=self.export_sweep_variable(dc.var),
            sweep=self.export_sweep(dc.sweep),
            ctrls=[],  # FIXME: analysis-specific controls
        )

    def export_ac(self, ac: data.Ac) -> vsp.AcInput:
        """Export an AC analysis"""
        analysis_name = ac.name or self.next_analysis_name()
        return vsp.AcInput(
            analysis_name=analysis_name,
            fstart=export_float(ac.sweep.start),
            fstop=export_float(ac.sweep.stop),
            npts=ac.sweep.npts,
            ctrls=[],  # FIXME: analysis-specific controls
        )

    def export_tran(self, tran: data.Tran) -> vsp.TranInput:
        """Export a transient analysis"""
        analysis_name = tran.name or self.next_analysis_name()
        return vsp.TranInput(
            analysis_name=analysis_name,
            # FIXME: VLSIR #26 move schema to Param / Prefixed
            tstop=export_float(tran.tstop),
            tstep=export_float(tran.tstep),
            ic={},  # FIXME: initial conditions
            ctrls=[],  # FIXME: analysis-specific controls
        )

    def export_noise(self, noise: data.Noise) -> vsp.NoiseInput:
        """Export a Noise analysis"""
        from ..instance import Instance
        from ..bundle import BundleInstance
        from ..diff_pair import Diff

        analysis_name = noise.name or self.next_analysis_name()

        # Sort out the output
        output = noise.output
        if isinstance(output, tuple):
            if len(output) != 2:
                raise ValueError(f"Invalid Noise Output: {output}")
            if any(not is_connectable for x in output):
                raise ValueError(f"Invalid Noise Output: {output}")
            output_p, output_n = output[0].name, output[1].name

        elif isinstance(output, BundleInstance):
            # Allow for `Diff` bundles
            if output.bundle is not Diff:
                raise ValueError(f"Invalid Noise Output: {output}")
            output_p, output_n = output.p.name, output.n.name

        elif is_connectable(output):
            # Single-ended Signal output
            output_p, output_n = output.name, ""

        elif isinstance(output, str):
            # Single-ended Signal output
            output_p, output_n = output, ""

        else:
            raise TypeError(f"Invalid Noise Output: {output}")

        # Sort out the input source
        if isinstance(noise.input_source, Instance):
            input_source = noise.input_source.name
        elif isinstance(noise.input_source, str):
            input_source = noise.input_source
        else:
            raise TypeError(f"Invalid Noise Input Source: {noise.input_source}")

        return vsp.NoiseInput(
            analysis_name=analysis_name,
            output_p=output_p,
            output_n=output_n,
            input_source=input_source,
            fstart=export_float(noise.sweep.start),
            fstop=export_float(noise.sweep.stop),
            npts=noise.sweep.npts,
            ctrls=[],  # FIXME: analysis-specific controls
        )

    def export_sweep_analysis(self, swp_an: data.SweepAnalysis) -> vsp.SweepInput:
        """Export a swept, nested set of one or more inner analyses as a `SweepInput`."""
        analysis_name = swp_an.name or self.next_analysis_name()
        return vsp.SweepInput(
            analysis_name=analysis_name,
            variable=self.export_sweep_variable(swp_an.var),
            sweep=self.export_sweep(swp_an.sweep),
            an=[self.export_analysis(a) for a in swp_an.inner],
            ctrls=[],  # FIXME: analysis-specific controls
        )

    def export_monte(self, monte: data.MonteCarlo) -> vsp.MonteInput:
        """Export a monte-carlo analysis"""
        analysis_name = monte.name or self.next_analysis_name()
        return vsp.MonteInput(
            analysis_name=analysis_name,
            npts=monte.npts,
            seed=0,  # FIXME: programmable seeds?
            an=[self.export_analysis(a) for a in monte.inner],
            ctrls=[],  # FIXME: analysis-specific controls
        )

    def export_custom_analysis(
        self, an: data.CustomAnalysis
    ) -> vsp.CustomAnalysisInput:
        """Export a custom analysis"""
        analysis_name = an.name or self.next_analysis_name()
        return vsp.CustomAnalysisInput(
            analysis_name=analysis_name,
            cmd=an.cmd,
            ctrls=[],  # FIXME: analysis-specific controls
        )

    def export_sweep_variable(self, var: Union[str, data.Param]) -> str:
        """Export a sweep-variable to its name."""
        if isinstance(var, str):
            return var
        elif isinstance(var, data.Param):
            return var.name
        raise TypeError(f"Invalid sweep variable {var}")

    def export_sweep(self, sweep: data.Sweep) -> vsp.Sweep:
        """Export a data sweep."""
        if isinstance(sweep, data.LinearSweep):
            return vsp.Sweep(
                linear=vsp.LinearSweep(
                    start=export_float(sweep.start),
                    stop=export_float(sweep.stop),
                    step=export_float(sweep.step),
                )
            )
        elif isinstance(sweep, data.LogSweep):
            return vsp.Sweep(
                log=vsp.LogSweep(
                    # FIXME: VLSIR #26 move schema to Param / Prefixed
                    start=export_float(sweep.start),
                    stop=export_float(sweep.stop),
                    npts=export_float(sweep.npts),  # FIXME: move to int
                )
            )
        elif isinstance(sweep, data.PointSweep):
            return vsp.Sweep(
                points=vsp.PointSweep(points=[export_float(x) for x in sweep.points])
            )
        else:
            raise TypeError(f"Invalid Sweep value {sweep}")


def export_options(options: data.Options) -> vsp.SimOptions:
    """Export simulation options"""
    from ..proto import export_param_value

    return vsp.SimOptions(name=options.name, value=export_param_value(options.value))


def export_control(ctrl: data.Control) -> vsp.Control:
    """Export a `Control` element"""
    from ..proto import export_literal

    if isinstance(ctrl, data.Include):
        return vsp.Control(include=export_include(ctrl))
    if isinstance(ctrl, data.Lib):
        return vsp.Control(lib=export_lib(ctrl))
    if isinstance(ctrl, data.Save):
        return vsp.Control(save=export_save(ctrl))
    if isinstance(ctrl, data.Meas):
        return vsp.Control(meas=export_meas(ctrl))
    if isinstance(ctrl, data.Param):
        return vsp.Control(param=export_param(ctrl))
    if isinstance(ctrl, Literal):
        return vsp.Control(literal=export_literal(ctrl))
    raise TypeError(f"Invalid Sim Control {ctrl}")


def export_include(inc: data.Include) -> vsp.Include:
    return vsp.Include(path=str(inc.path))


def export_lib(lib: data.Lib) -> vsp.LibInclude:
    return vsp.LibInclude(path=str(lib.path), section=lib.section)


def export_save(save: data.Save) -> vsp.Save:
    if isinstance(save.targ, data.SaveMode):
        if save.targ == data.SaveMode.ALL:
            mode = vsp.Save.SaveMode.ALL
        elif save.targ == data.SaveMode.NONE:
            mode = vsp.Save.SaveMode.NONE
        else:
            raise ValueError
        return vsp.Save(mode=mode)
    if isinstance(save.targ, Signal):
        signal = save.targ.name
    elif isinstance(save.targ, List[Signal]):
        signal = ",".join([s.name for s in save.targ])
    elif isinstance(save.targ, str):
        signal = save.targ
    elif isinstance(save.targ, List[str]):
        signal = ",".join([s for s in save.targ])
    else:
        raise TypeError
    return vsp.Save(signal=signal)


def export_meas(meas: data.Meas) -> vsp.Meas:
    """Export a measurement"""
    return vsp.Meas(
        analysis_type=export_analysis_type(meas.analysis),
        name=meas.name,
        expr=str(meas.expr),
    )


def export_analysis_type(an: Union[str, data.Analysis]) -> str:
    """Export an `AnalysisType`, or string representation thereof."""
    if isinstance(an, str):
        return an
    if data.is_analysis(an):
        return an.tp.value
    raise TypeError(f"Invalid Analysis for type-extraction {an}")


def export_param(param: data.Param) -> vlsir.Param:
    """Export a parameter declaration"""
    from ..proto import export_param_value

    return vlsir.Param(name=param.name, value=export_param_value(param.val))


def export_float(num: Union[float, int, Decimal, Prefixed, Scalar]) -> float:
    """Export a `Number` union-type to a float, or protobuf float/double."""
    if num is None:
        # FIXME: this is the protobuf default, but we really wanna make fields that use it optional
        return 0.0
    if isinstance(num, float):
        return num
    if isinstance(num, (int, str, Decimal, Prefixed)):
        return float(num)
    raise TypeError(f"Invalid value for proto float: {num}")



""" 
# Hdl21 Unit Tests 
"""

from typing import TypeVar
import copy, pytest
import hdl21 as h

T = TypeVar("T")


def test_version():
    assert h.__version__ == "7.0.0"  # VLSIR_VERSION


def test_module1():
    """Initial Module Test"""

    @h.module
    class M1:
        a = h.Input()
        b = h.Output()
        c = h.Inout()
        d = h.Port()
        e = h.Signal()
        f = h.Signal()

    assert isinstance(M1, h.Module)
    assert isinstance(M1.ports, dict)
    assert isinstance(M1.signals, dict)
    assert isinstance(M1.instances, dict)
    assert "a" in M1.ports
    assert "a" not in M1.signals
    assert "b" in M1.ports
    assert "b" not in M1.signals
    assert "c" in M1.ports
    assert "c" not in M1.signals
    assert "d" in M1.ports
    assert "d" not in M1.signals
    assert "e" in M1.signals
    assert "e" not in M1.ports
    assert "f" in M1.signals
    assert "f" not in M1.ports
    assert M1.e is not M1.f


def test_module2():
    @h.module
    class M1:
        s = h.Input()

    @h.module
    class M2:
        q = h.Signal()
        i = M1(s=q)

    assert isinstance(M1, h.Module)
    assert isinstance(M2, h.Module)
    assert isinstance(M1.ports["s"], h.Signal)
    assert isinstance(M2.signals["q"], h.Signal)
    assert isinstance(M2.instances["i"], h.Instance)

    # Test the getattr magic for signals, ports, and instances
    assert isinstance(M2.i, h.Instance)
    assert M2.instances["i"] is M2.i
    assert isinstance(M2.q, h.Signal)
    assert M2.signals["q"] is M2.q
    assert isinstance(M1.s, h.Signal)
    assert M1.ports["s"] is M1.s


def test_generator1():
    @h.paramclass
    class MyParams:
        w = h.Param(dtype=int, desc="five", default=5)

    @h.generator
    def gen1(params: MyParams) -> h.Module:
        m = h.Module()
        m.i = h.Input(width=params.w)
        return m

    m = h.elaborate(gen1(MyParams(w=3)))

    assert m.name == "gen1(w=3)"
    assert isinstance(m.i, h.Signal)


@pytest.mark.xfail(reason="Deprecated/ unimplemented Context feature")
def test_generator2():
    @h.paramclass
    class P2:
        f = h.Param(dtype=float, desc="a real number", default=1e-11)

    @h.generator
    def g2(params: P2) -> h.Module:
        assert isinstance(params, P2)
        return h.Module()

    m = h.elaborate(g2(P2()))

    assert isinstance(m, h.Module)
    assert m.name == "g2(f=1e-11)"


def test_generator3():
    @h.paramclass
    class P3:
        width = h.Param(dtype=int, desc="bit-width, maybe", default=1)

    @h.generator
    def g3a(params: P3) -> h.Module:
        return h.Module()

    @h.generator
    def g3b(params: P3) -> h.Module:
        return h.Module()

    M = h.Module(name="M")
    p3a = P3()
    p3b = P3(width=5)

    @h.module
    class HasGen:
        a = g3a(p3a)()
        b = g3b(p3b)()
        c = M()

    # Elaborate the top module
    h.elaborate(HasGen)

    # Post-elab checks
    from hdl21.generator import GeneratorCall

    assert isinstance(HasGen.a, h.Instance)
    call = HasGen.a.of._generated_by
    assert isinstance(call, GeneratorCall)
    assert call.gen is g3a
    assert call.params == P3()
    result = h.Generator.Cache.done.get(call)
    assert isinstance(result, h.Module)
    assert result.name == "g3a(width=1)"
    assert isinstance(HasGen.b, h.Instance)
    call = HasGen.b.of._generated_by
    assert isinstance(call, GeneratorCall)
    result = h.Generator.Cache.done.get(call)
    assert isinstance(result, h.Module)
    assert result.name == "g3b(width=5)"
    assert call.gen is g3b
    assert call.params == P3(width=5)
    assert isinstance(HasGen.c, h.Instance)
    assert isinstance(HasGen.c.of, h.Module)
    assert HasGen.c.of is M


def test_array1():
    @h.module
    class InArray:
        inp = h.Input()
        out = h.Output()

    m = h.Module(name="HasArray")
    m.s1 = h.Signal(width=8)
    m.s2 = h.Signal(width=1)
    m.arr = h.InstanceArray(InArray, 8)
    m.arr.inp = m.s1
    m.arr.out = m.s2

    assert m.name == "HasArray"


def test_array2():
    """Basic Instance-Array Test"""
    a = h.Module(name="a")
    a.inp = h.Port(width=1)
    a.out = h.Port(width=1)

    @h.module
    class HasArray2:
        s1 = h.Signal(width=8)
        s2 = h.Signal(width=1)
        arr = h.InstanceArray(a, 8)(inp=s1, out=s2)

    assert len(HasArray2.instances) == 0
    assert len(HasArray2.instarrays) == 1

    # Elaborate, flattening arrays along the way
    h.elaborate(HasArray2)

    # Post-elab checks
    assert len(HasArray2.instances) == 8
    assert len(HasArray2.instarrays) == 0


def test_cycle1():
    """Test cyclical connection-graphs, i.e. a back-to-back pair of instances"""

    @h.module
    class Thing:
        inp = h.Input()
        out = h.Output()

    @h.module
    class BackToBack:
        t1 = Thing()
        t2 = Thing(inp=t1.out, out=t1.inp)

    assert isinstance(BackToBack.t1.inp, h.PortRef)
    assert isinstance(BackToBack.t1.out, h.PortRef)
    assert isinstance(BackToBack.t2.inp, h.PortRef)
    assert isinstance(BackToBack.t2.out, h.PortRef)

    b = h.elaborate(BackToBack)

    assert isinstance(b.t1.inp, h.Signal)
    assert isinstance(b.t1.out, h.Signal)
    assert isinstance(b.t2.inp, h.Signal)
    assert isinstance(b.t2.out, h.Signal)


def test_cycle2():
    """# Doing the same thing in procedural code"""

    @h.module
    class Thing:
        inp = h.Input()
        out = h.Output()

    b2 = h.Module(name="BackToBack2")
    b2.t1 = Thing()
    b2.t2 = Thing()
    b2.t2.inp = b2.t1.out
    b2.t2.out = b2.t1.inp

    assert isinstance(b2.t1.inp, h.PortRef)
    assert isinstance(b2.t1.out, h.PortRef)
    assert isinstance(b2.t2.inp, h.PortRef)
    assert isinstance(b2.t2.out, h.PortRef)

    b2 = h.elaborate(b2)

    assert len(b2.instances) == 2
    assert len(b2.instarrays) == 0
    assert len(b2.bundles) == 0
    assert len(b2.ports) == 0
    assert len(b2.signals) == 2
    assert "t1_out" in b2.signals
    assert "t1_inp" in b2.signals


def test_gen3():
    @h.paramclass
    class MyParams:
        w = h.Param(dtype=int, desc="Input bit-width. Required")

    @h.generator
    def MyThirdGenerator(params: MyParams) -> h.Module:
        @h.module
        class Inner:
            i = h.Input(width=params.w)
            o = h.Output(width=2 * params.w)

        # Instantiate that in another Module
        @h.module
        class Outer:
            i = h.Signal(width=params.w)
            o = h.Signal(width=2 * params.w)
            inner = Inner(i=i, o=o)

        # And manipulate that some more too
        Outer.inp = h.Input(width=params.w)
        return Outer

    h.elaborate(MyThirdGenerator(MyParams(1)))

    # FIXME: post-elab checks


def test_prim1():
    # First test of transistor primitives
    params = h.Mos.Params()
    pmos = h.Pmos(params)
    nmos = h.Nmos(params)

    @h.module
    class HasMos:
        # Two transistors wired in parallel
        p = pmos()
        n = nmos(g=p.g, d=p.d, s=p.s, b=p.b)

    h.elaborate(HasMos)
    # FIXME: post-elab checks


def test_prim2():
    @h.module
    class HasPrims:
        p = h.Signal()
        n = h.Signal()

        _rp = h.R.Params(r=50)
        r = h.Resistor(_rp)(p=p, n=n)

        c = h.Capacitor(h.C.Params(c=1e-12))(p=p, n=n)
        l = h.Inductor(h.L.Params(l=1e-15))(p=p, n=n)
        d = h.Diode(h.D.Params())(p=p, n=n)
        s = h.Short(h.Short.Params())(p=p, n=n)

    h.elaborate(HasPrims)
    # FIXME: post-elab checks


def test_signal_slice1():
    # Initial test of signal slicing

    sig = h.Signal(width=10)
    sl = sig[0]
    assert isinstance(sl, h.Slice)
    assert sl.top == 1
    assert sl.bot == 0
    assert sl.width == 1
    assert sl.step == 1
    assert sl.parent is sig

    sl = sig[0:5]
    assert isinstance(sl, h.Slice)
    assert sl.top == 5
    assert sl.bot == 0
    assert sl.width == 5
    assert sl.step == 1
    assert sl.parent is sig

    sl = sig[:]
    assert isinstance(sl, h.Slice)
    assert sl.top == 10
    assert sl.bot == 0
    assert sl.width == 10
    assert sl.step == 1
    assert sl.parent is sig


def test_signal_slice2():
    # Test slicing advanced features
    sl = h.Signal(width=11)[-1]
    assert sl.top == 11
    assert sl.bot == 10
    assert sl.step == 1
    assert sl.width == 1

    sl = h.Signal(width=11)[:-1]
    assert sl.top == 10
    assert sl.bot == 0
    assert sl.step == 1
    assert sl.width == 10

    sl = h.Signal(width=11)[-2:]
    assert sl.top == 11
    assert sl.bot == 9
    assert sl.step == 1
    assert sl.width == 2


@pytest.mark.xfail(reason="#21 https://github.com/dan-fritchman/Hdl21/issues/21")
def test_bad_slice1():
    # Test slicing error-cases
    with pytest.raises(TypeError):
        h.Signal(width=11)[None]

    with pytest.raises(ValueError):
        h.Signal(width=11)[11]

    h.Signal(width=11)[2:1:-1]  # OK
    with pytest.raises(ValueError):
        h.Signal(width=11)[2:1:1]  # Not OK

    h.Signal(width=11)[1:9]  # OK
    with pytest.raises(ValueError):
        h.Signal(width=11)[9:1]  # Not OK


def test_signal_concat1():
    # Initial Signal concatenation test

    c = h.Concat(h.Signal(width=1), h.Signal(width=2), h.Signal(width=3))
    assert isinstance(c, h.Concat)
    assert len(c.parts) == 3
    for p in c.parts:
        assert isinstance(p, h.Signal)
    assert c.width == 6

    c = h.Concat(h.Signal(width=1)[0], h.Signal(width=2)[0], h.Signal(width=3)[0])
    assert isinstance(c, h.Concat)
    assert len(c.parts) == 3
    for p in c.parts:
        assert isinstance(p, h.Slice)
    assert c.width == 3

    c = h.Concat(
        h.Concat(h.Signal(), h.Signal()), h.Signal(width=2), h.Signal(width=2)[:]
    )
    assert isinstance(c, h.Concat)
    assert len(c.parts) == 3
    assert c.width == 6


def test_slice_module1():
    """Make use of slicing and concatenation in a module"""

    C = h.Module(name="C")
    C.p1 = h.Port(width=1)
    C.p4 = h.Port(width=4)
    C.p7 = h.Port(width=7)

    P = h.Module(name="P")
    P.s4 = h.Signal(width=4)
    P.s2 = h.Signal(width=2)
    P.ic = C(p1=P.s4[0], p4=P.s4, p7=h.Concat(P.s4, P.s2, P.s2[0]))

    h.elaborate(P)


def test_instantiable_as_param():
    """Test using each Instantiable type as a parameter-value."""

    @h.paramclass
    class Params:
        m = h.Param(dtype=h.Module, desc="A `Module`")
        e = h.Param(dtype=h.ExternalModule, desc="An `ExternalModule`")
        ec = h.Param(dtype=h.ExternalModuleCall, desc="An `ExternalModuleCall`")
        g = h.Param(dtype=h.Generator, desc="A `Generator`")
        p = h.Param(dtype=h.Primitive, desc="A `Primitive`")
        pc = h.Param(dtype=h.PrimitiveCall, desc="A `PrimitiveCall`")
        i = h.Param(dtype=h.Instantiable, desc="An `Instantiable`")

    @h.generator
    def UsesThemParams(params: Params) -> h.Module:
        @h.module
        class M:
            x = h.Signal()

            m = params.m(x=x)
            e = params.e()(x=x)
            ec = params.ec(x=x)
            g = params.g()(x=x)
            p = params.p()(d=x, g=x, s=x, b=x)
            pc = params.pc(d=x, g=x, s=x, b=x)
            i = params.i(x=x)

        return M

    Mod = h.Module(name="Mod")
    Mod.x = h.Port()
    Emod = h.ExternalModule(name="Emod", port_list=[h.Port(name="x")])

    @h.generator
    def Gen(_: h.HasNoParams) -> h.Module:
        m = h.Module()
        m.x = h.Port()
        return m

    p = Params(
        m=Mod,
        e=Emod,
        ec=Emod(),
        g=Gen,
        p=h.Mos,
        pc=h.Mos(),
        i=Mod,
    )
    m = UsesThemParams(p)
    m = h.elaborate(m)


def test_instance_mult():
    """Initial tests of instance-array-generation by multiplication"""

    Child = h.Module(name="Child")
    Parent = h.Module(name="Parent")

    # Create an array via multiplication
    Parent.child = Child() * 5  # Lotta kids here

    # Check that the array is created correctly
    assert isinstance(Parent.child, h.InstanceArray)
    assert Parent.child.n == 5
    assert Parent.child.of is Child

    # Test elaboration
    h.elaborate(Parent)

    # Post-elaboration checks
    assert len(Parent.instances) == 5
    assert len(Parent.instarrays) == 0
    for inst in Parent.instances.values():
        assert inst.of is Child

    # Create another, this time via right-mul
    Parent = h.Module(name="Parent")
    Parent.child = 11 * Child()  # Getting even more kids

    # Check that the array is created correctly
    assert isinstance(Parent.child, h.InstanceArray)
    assert Parent.child.n == 11
    assert Parent.child.of is Child

    # Test elaboration
    h.elaborate(Parent)

    # Post-elaboration checks
    assert len(Parent.instances) == 11
    assert len(Parent.instarrays) == 0
    for inst in Parent.instances.values():
        assert inst.of is Child


def test_instance_mult2():
    """Test connecting to multiplication-generated InstanceArrays"""

    @h.module
    class Child:
        p = h.Port()

    @h.module
    class Parent:
        a = h.Signal(width=3)
        child = (Child() * 3)(p=a)

    assert len(Parent.instances) == 0
    assert len(Parent.instarrays) == 1

    h.elaborate(Parent)

    # Check that array-flattening completed correctly
    assert len(Parent.instances) == 3
    assert len(Parent.instarrays) == 0
    assert Parent.get("child_0").of is Child
    assert Parent.get("child_1").of is Child
    assert Parent.get("child_2").of is Child

    # FIXME: slice equality changes break these tests
    # assert Parent.get("child_0").conns["p"] == Parent.a[0]
    # assert Parent.get("child_1").conns["p"] == Parent.a[1]
    # assert Parent.get("child_2").conns["p"] == Parent.a[2]


def test_instance_mult3():
    """Test connecting to an "already connected" Instance"""

    @h.module
    class Child:
        p = h.Port()

    @h.module
    class Parent:
        a = h.Signal(width=3)
        child = 3 * Child(p=a)  # <= this the one diff here from test_instance_mult2

    h.elaborate(Parent)

    # Check that array-flattening completed correctly
    assert len(Parent.instances) == 3
    assert len(Parent.instarrays) == 0
    assert Parent.get("child_0").of is Child
    assert Parent.get("child_1").of is Child
    assert Parent.get("child_2").of is Child

    # FIXME: slice equality changes break these tests
    # assert Parent.get("child_0").conns["p"] == Parent.a[0]
    # assert Parent.get("child_1").conns["p"] == Parent.a[1]
    # assert Parent.get("child_2").conns["p"] == Parent.a[2]


def test_instance_mult4():
    """Test connecting non-unit-width Arrays"""

    @h.module
    class Child:
        a = h.Port(width=11)
        b = h.Port(width=16)

    @h.module
    class Parent:
        a = h.Signal(width=11)
        b = h.Signal(width=48)
        child = 3 * Child(a=a, b=b)

    h.elaborate(Parent)

    # Check that array-flattening completed correctly
    assert len(Parent.instances) == 3
    assert len(Parent.instarrays) == 0
    assert Parent.get("child_0").of is Child
    assert Parent.get("child_1").of is Child
    assert Parent.get("child_2").of is Child
    assert Parent.get("child_0").conns["a"] == Parent.a
    assert Parent.get("child_1").conns["a"] == Parent.a
    assert Parent.get("child_2").conns["a"] == Parent.a

    # FIXME: we probably want a new Signal/ Slice "equality" method for these
    # assert Parent.get("child_0").conns["b"] == Parent.b[0:16]
    # assert Parent.get("child_1").conns["b"] == Parent.b[16:32]
    # assert Parent.get("child_2").conns["b"] == Parent.b[32:48]


def test_bad_width_conn():
    """Test invalid connection-widths"""
    c = h.Module(name="c")
    c.p = h.Port(width=3)  # Width-3 Port
    q = h.Module(name="q")
    q.s = h.Signal(width=5)  # Width-5 Signal
    q.c = c(p=q.s)  # <= Bad connection here

    with pytest.raises(RuntimeError):
        h.elaborate(q)


def test_illegal_module_attrs():
    """Test attempting to add illegal attributes"""

    m = h.Module()
    with pytest.raises(TypeError):
        m.a = list()

    @h.module
    class M:
        # This version works as of v2.0, since it makes `a`
        # easier to refer to, e.g. as part of an instance parameter.
        a = list()

    @h.module
    class C:
        ...

    with pytest.raises(TypeError):
        C.b = TabError

    # Use combinations of legal attributes

    m = h.Module(name="m")
    with pytest.raises(TypeError):
        m.p = 5
    with pytest.raises(TypeError):
        m.z = dict(a=h.Signal())
    with pytest.raises(TypeError):
        m.q = [h.Port()]
    with pytest.raises(TypeError):
        m.p = (h.Input(), h.Output())
    with pytest.raises(TypeError):
        m.func = lambda _: None

    # As of v2.0 this is allowed, although `fail` does not show up on the resultant `Module`.
    @h.module
    class HasMethod:
        def fail(self):
            ...

    assert not hasattr(HasMethod, "fail")


def test_copy_signal():
    """Copying a Signal"""
    copy.copy(h.Signal())
    copy.deepcopy(h.Signal())


def test_orphanage():
    """Test that orphaned Module-attributes fail at elaboration"""

    m1 = h.Module(name="m1")
    m1.s = h.Signal()  # Signal `s` is now "parented" by `m1`

    m2 = h.Module(name="m2")
    m2.y = m1.s  # Now `s` has been "orphaned" (or perhaps "cradle-robbed") by `m2`

    # Note elaborating `m2` continues to work
    h.elaborate(m2)

    with pytest.raises(RuntimeError):
        # Elaborating `m1` should fail, since `s` is orphaned
        h.elaborate(m1)


def test_orphanage2():
    """Test orphaning a bundle instance"""

    @h.bundle
    class B:
        bb = h.Signal(width=1)

    b = B()  # Instance of `B`-type Bundle, to be orphaned

    m1 = h.Module(name="m1")
    m1.b = b
    m2 = h.Module(name="m2")
    m2.b = b

    # Note elaborating `m2` continues to work
    h.elaborate(m2)

    with pytest.raises(RuntimeError):
        # Elaborating `m1` should fail, since `s` is orphaned
        h.elaborate(m1)


def test_orphanage3():
    """Test orphaning an instance"""
    I = h.Module(name="I")
    i = I()  # Instance to be orphaned

    m1 = h.Module(name="m1")
    m1.i = i
    m2 = h.Module(name="m2")
    m2.i = i

    # Note elaborating `m2` continues to work
    h.elaborate(m2)

    with pytest.raises(RuntimeError):
        # Elaborating `m1` should fail, since `s` is orphaned
        h.elaborate(m1)


def test_orphanage4():
    """Test an orphan Instance connection"""

    m1 = h.Module(name="m1")
    m1.p = h.Port()

    # The "orphan signal" will not be owned by any module
    the_orphan_signal = h.Signal()
    m2 = h.Module(name="m2")
    m2.i = m1(p=the_orphan_signal)  # <= Problem's here

    # Note elaborating `m1` continues to work
    h.elaborate(m1)

    with pytest.raises(RuntimeError):
        # Elaborating `m2` should fail
        h.elaborate(m2)


def test_wrong_decorator():
    """Mistake `Module` for `module` and vice versa"""

    with pytest.raises(TypeError) as e:

        @h.Module  # Bad!
        class M:
            ...

    assert "Did you mean to use the `module` decorator?" in str(e)

    with pytest.raises(TypeError) as e:
        h.Module(2)  # Bad!

    assert "Invalid Module name" in str(e)

    with pytest.raises(TypeError) as e:
        ok = h.Module("ok")  # OK
        h.Module(ok)  # Bad!

    assert "Invalid Module name" in str(e)


def test_elab_noconn():
    """Initial test of elaborating a `NoConn`"""

    @h.module
    class Inner:
        p = h.Port()

    @h.module
    class HasNoConn:
        i1 = Inner()
        i1.p = h.NoConn()

    h.elaborate(HasNoConn)

    assert len(HasNoConn.signals) == 1


def test_elab_noconn2():
    """Slightly more elaborate test of elaborating a `NoConn`"""

    @h.module
    class M3:  # Layer 3: just a Port
        p = h.Port()

    @h.module
    class M2:  # Layer 2: instantiate & connect M3
        p = h.Port()
        m3 = M3(p=p)

    @h.module
    class M1:  # Layer 1: connect a `NoConn` to all that
        p = h.NoConn()
        m2 = M2(p=p)

    h.elaborate(M1)

    assert len(M1.signals) == 1


def test_bad_noconn():
    """Test that a doubly-connected `NoConn` should fail"""

    @h.module
    class Inner:
        p = h.Port()

    @h.module
    class Bad:
        # Create two instances of `Inner`
        i1 = Inner()
        i2 = Inner()

        # Connect the first to a `NoConn`
        i1.p = h.NoConn()
        # And then connect the second to the first.
        # Herein lies the "problem"
        i2.p = i1.p

    with pytest.raises(RuntimeError):
        h.elaborate(Bad)


def test_array_concat_conn():
    """Test connecting a `Concat` to an `InstanceArray`"""

    Child = h.Module(name="Child")
    Child.p3 = h.Input(width=3)
    Child.p5 = h.Input(width=5)

    Parent = h.Module(name="Parent")
    Parent.s5 = h.Signal(width=5)
    Parent.s9 = h.Signal(width=9)
    Parent.c = 2 * Child()
    Parent.c.p3 = Parent.s5[0:3]
    Parent.c.p5 = h.Concat(Parent.s5[3], Parent.s9)


@pytest.mark.xfail(reason="FIXME: structural Slice/ Concat equality ")
def test_slice_resolution():
    """Test resolutions of slice combinations"""

    # This is a very private import of the slice-resolver function
    from hdl21.elab.elaborators.slices import _resolve_slice

    # Slice of a Signal
    s = h.Signal(width=5)
    assert _resolve_slice(s[0]) == s[0]

    # Slice of a Concat
    s1 = h.Signal(name="s1", width=5)
    s2 = h.Signal(name="s2", width=5)
    c = h.Concat(s1, s2)
    assert _resolve_slice(c[0]) == s1[0]
    assert _resolve_slice(c[1]) == s1[1]
    assert _resolve_slice(c[2]) == s1[2]
    assert _resolve_slice(c[3]) == s1[3]
    assert _resolve_slice(c[4]) == s1[4]
    assert _resolve_slice(c[5]) == s2[0]
    assert _resolve_slice(c[6]) == s2[1]
    assert _resolve_slice(c[7]) == s2[2]
    assert _resolve_slice(c[8]) == s2[3]
    assert _resolve_slice(c[9]) == s2[4]

    # Slice of a Slice
    sa = h.Signal(name="sa", width=5)
    assert _resolve_slice(sa[2:5][0]) == sa[2]
    assert _resolve_slice(sa[2:5][1]) == sa[3]
    assert _resolve_slice(sa[2:5][2]) == sa[4]

    assert _resolve_slice(sa[:][0]) == sa[0]
    assert _resolve_slice(sa[:][1]) == sa[1]
    assert _resolve_slice(sa[:][2]) == sa[2]
    assert _resolve_slice(sa[:][3]) == sa[3]
    assert _resolve_slice(sa[:][4]) == sa[4]

    # Check a concatenation case. Note `Concat` does not support equality, so compare parts.
    assert _resolve_slice(sa[:][0:4]).parts == (sa[0], sa[1], sa[2], sa[3])


def test_common_attr_errors():
    """Test that common errors provide helpful feedback.
    For example, adding `Module`s where one wants an `Instance`."""

    M = h.Module(name="M")

    # Module assignment, where we'd want an Instance `M.i = I()`
    I = h.Module(name="I")
    with pytest.raises(TypeError) as einfo:
        M.i = I  # Bad - Module
    assert "Did you mean" in str(einfo.value)
    assert "`Module`" in str(einfo.value)
    M.i = I()  # Good - Instance

    # Primitive assignment
    from hdl21.primitives import R

    with pytest.raises(TypeError) as einfo:
        M.r = R  # Fail - Primitive
    assert "Did you mean" in str(einfo.value)
    assert "`Primitive`" in str(einfo.value)
    with pytest.raises(TypeError) as einfo:
        M.r = R(R.Params(r=1))  # Fail - Primitive Call
    assert "Did you mean" in str(einfo.value)
    assert "`PrimitiveCall`" in str(einfo.value)
    M.r = R(R.Params(r=1))()  # Good - Instances (granted incompletely connected)

    # Generator assignment
    @h.generator
    def G(p: h.HasNoParams) -> h.Module:
        return h.Module()

    with pytest.raises(TypeError) as einfo:
        M.g = G  # Bad - Generator
    assert "Did you mean" in str(einfo.value)
    assert "`Generator`" in str(einfo.value)
    with pytest.raises(TypeError) as einfo:
        M.g = G(h.NoParams)  # Bad - GeneratorCall
    assert "Did you mean" in str(einfo.value)
    assert "`Module`" in str(einfo.value)
    M.g = G(h.NoParams)()  # Good - Instance

    X = h.ExternalModule(name="X", port_list=[])
    with pytest.raises(TypeError) as einfo:
        M.x = X  # Bad - ExternalModule
    assert "Did you mean" in str(einfo.value)
    assert "`ExternalModule`" in str(einfo.value)
    with pytest.raises(TypeError) as einfo:
        M.x = X()  # Bad - ExternalModuleCall
    assert "Did you mean" in str(einfo.value)
    assert "`ExternalModuleCall`" in str(einfo.value)
    M.x = X()()  # Good - Instance


def test_invalid_instantiable_error():
    """# Test the errors for common invalid `Instantiable`s
    Inspired by issue #174."""

    @h.paramclass
    class P:  # Param class with an `Instantiable` field
        i = h.Param(dtype=h.Instantiable, desc="Module to instantiate")

    @h.generator
    def G(_: h.HasNoParams) -> h.Module:
        return h.Module()

    def ok(val: T) -> T:
        P(i=val)
        return val

    def bad(val: T) -> T:
        with pytest.raises(Exception) as einfo:
            P(i=val)
        assert "Did you mean" in str(einfo.value)
        return val

    # Create parameters for each of the above
    M = ok(h.Module(name="M"))  # Module
    _ = bad(M())  # Instance

    X = bad(h.ExternalModule(name="X", port_list=[]))
    XC = ok(X())  # ExternalModuleCall
    _ = bad(XC())  # Instance of ExternalModuleCall

    R = bad(h.primitives.Res)  # Primitive
    RC = ok(R(r=1))  # PrimitiveCall
    _ = bad(RC())  # Instance of PrimitiveCall

    GC = ok(G())  # Generated Module
    _ = bad(GC())  # Instance of generated Module


def test_generator_call_by_kwargs():
    """Test the capacity for generators to create their param-classes inline."""

    @h.paramclass
    class P:
        a = h.Param(dtype=int, desc="a")
        b = h.Param(dtype=float, desc="b")
        c = h.Param(dtype=str, desc="c")

    @h.generator
    def M(_: P) -> h.Module:
        return h.Module()

    # Call without constructing a `P`
    m = M(a=1, b=2.0, c="3")
    call = m._generated_by

    from hdl21.generator import GeneratorCall

    assert isinstance(call, GeneratorCall)
    m = h.elaborate(m)
    assert isinstance(m, h.Module)

    # Check that type-checking continue
    with pytest.raises(RuntimeError):
        m = M(a=TabError, b=TabError, c=TabError)


def test_instance_array_portrefs():
    """Test Instance Arrays connected by port-references"""

    Inv = h.ExternalModule(
        name="Inv",
        port_list=[
            h.Input(name="i"),
            h.Output(name="z"),
        ],
    )

    m = h.Module(name="TestArrayPortRef")
    m.a, m.b = h.Signals(2, width=4)

    # Create an InstanceArray
    m.inva = 4 * Inv()(i=m.a)
    # And another which connects to it via PortRef
    m.invb = 4 * Inv()(i=m.inva.z, z=m.b)

    h.elaborate(m)

    assert len(m.instances) == 8
    assert len(m.instarrays) == 0


def test_array_bundle():
    """Test bundle-valued connections to instance arrays."""

    @h.bundle
    class B:
        s = h.Signal()

    @h.module
    class HasB:
        b = B(port=True)

    @h.module
    class HasArr:
        b = B()
        bs = 11 * HasB(b=b)

    # Elaborate this,
    h.elaborate(HasArr)

    assert len(HasArr.signals) == 1
    assert len(HasArr.instances) == 11
    assert len(HasArr.instarrays) == 0
    for inst in HasArr.instances.values():
        assert inst.conns.get("b_s", None) is HasArr.b_s


def test_multiple_signals_in_port_group():
    """Test, or at least try to test, jamming more than one Signal
    into the `group` concept used by the `ResolvePortRefs` elaborator pass.
    Based on the design of `ResolvePortRefs`, this is a thing that (we think)
    is not possible."""

    @h.module
    class I:
        s = h.Port()

    @h.module
    class M:
        s = h.Signal()
        i1 = I(s=s)
        i2 = I(s=s)
        i3 = I(s=i1.s)
        i4 = I(s=i2.s)

    h.elaborate(M)

    # Check this resolved to a single Signal
    assert len(M.signals) == 1
    assert len(M.bundles) == 0
    assert len(M.instances) == 4

    # And check that Signal is connected to all four Instances
    assert M.i1.s is M.s
    assert M.i2.s is M.s
    assert M.i3.s is M.s
    assert M.i4.s is M.s


def test_bad_conns():
    """Test the errors produced by elaborating with invalid Port connections"""

    @h.module
    class I:
        p = h.Port()

    @h.module
    class O:
        a, b, c = h.Signals(3)
        i = I(a=a, b=b, p=c)  # Two of these 3 ports don't exist

    with pytest.raises(RuntimeError) as e:
        h.elaborate(O)
    assert "Connection to non-existent Port" in str(e.value)

    @h.module
    class O:
        a, b, c = h.Signals(3)
        i = I(a=a, b=b)  # None of these ports exist

    with pytest.raises(RuntimeError) as e:
        h.elaborate(O)
    assert "Missing connection to" in str(e.value)
    assert "Connection to non-existent Port" in str(e.value)

    @h.module
    class O:
        a, b, c = h.Signals(3)
        i = I(a=a, p=b)  # One of these ports don't exist

    with pytest.raises(RuntimeError) as e:
        h.elaborate(O)
    assert "Connection to non-existent Port" in str(e.value)

    @h.module
    class O:
        a, b, c = h.Signals(3)
        i = I(a=a)  # All (one) port doesn't exist

    with pytest.raises(RuntimeError) as e:
        h.elaborate(O)
    assert "Missing connection to" in str(e.value)


def test_generator_port_slice():
    """Test taking a slice from a Generator `PortRef`."""

    @h.paramclass
    class Width:
        width = h.Param(dtype=int, desc="Width")

    @h.generator
    def SliceMyP(params: Width) -> h.Module:
        m = h.Module()
        m.p = h.Port(width=params.width)  # Port with parametrized width
        return m

    @h.module
    class ScalarPort:
        p = h.Port()  # A width-one port

    @h.module
    class HasGen:
        g = SliceMyP(width=8)()
        # Connect the MSB of `g.p` to `s.p`
        s = ScalarPort(p=g.p[-1])

    h.elaborate(HasGen)


def test_pair1():
    """Test of the `Diff` and `Pair` signal and instance bundles"""

    @h.module
    class I:
        ps = h.Port(desc="Port we'll connect to a scalar")
        pd = h.Port(desc="Port we'll connect to a diff")
        pa = h.Port(desc="Port we'll connect to an AnonymousBundle")

    @h.module
    class O:
        d = h.Diff()
        s = h.Signal()
        pair = h.Pair(I)(pd=d, ps=s, pa=h.inverse(d))

    h.elaborate(O)

    assert len(O.instances) == 2
    assert "pair_p" in O.instances
    assert "pair_n" in O.instances
    assert O.pair_p.conns["ps"] is O.s
    assert O.pair_n.conns["ps"] is O.s
    assert O.pair_p.conns["pd"] is O.d_p
    assert O.pair_n.conns["pd"] is O.d_n
    assert O.pair_p.conns["pa"] is O.d_n
    assert O.pair_n.conns["pa"] is O.d_p
    assert len(O.signals) == 3
    assert len(O.bundles) == 0
    assert len(O.instbundles) == 0


def test_noconn_types():
    """Test connecting `NoConn`s to a variety of port-types."""

    @h.module
    class Inner:
        # Inner module with a handful of port-types
        s = h.Port()
        b = h.Port(width=11)
        d = h.Diff(port=True)

    @h.module
    class Outer:
        i = Inner(s=h.NoConn(), b=h.NoConn(), d=h.NoConn())

    # Most of the test is here, just seeing that this can elaborate
    h.elaborate(Outer)

    assert len(Inner.signals) == 0
    assert len(Outer.ports) == 0
    assert sorted(Inner.ports.keys()) == ["b", "d_n", "d_p", "s"]
    assert sorted(Outer.signals.keys()) == ["i_b", "i_d_n", "i_d_p", "i_s"]
    assert Outer.i.conns["b"] is Outer.i_b
    assert Outer.i.conns["d_n"] is Outer.i_d_n
    assert Outer.i.conns["d_p"] is Outer.i_d_p
    assert Outer.i.conns["s"] is Outer.i_s


def test_deep_hierarchy():
    """Test a deep hierarchy, without much content."""

    # Create the leaf level module. Empty with a single Port `p`.
    M0 = h.Module(name=f"M0")
    M0.p = h.Port()
    prev = M0
    M1 = None

    # Create lots of layers above it, each of which instantiates the previous.
    for k in range(100):
        m = h.Module(name=f"M{k}")
        m.p = h.Port()
        m.i = h.Instance(of=prev)(p=m.p)
        if M1 is None:
            M1 = m
        prev = m

    # Screw up a connection, and check that we get an elaboration-time error when elaborating the top-level module.
    M1.i.not_a_real_port = M1.p
    with pytest.raises(RuntimeError):
        h.elaborate(m)


def test_generator_eq():
    """Test equality and hashing of Generator calls using `Default`."""

    @h.paramclass
    class Params:
        p = h.Param(dtype=int, desc="p", default=111)

    @h.generator
    def Gen(_: Params) -> h.Module:
        return h.Module()

    assert Gen() == Gen()
    assert Gen() == Gen(Params(p=111))
    assert Gen() == Gen(p=111)
    assert Gen()._generated_by.params == Params()
    assert Gen()._generated_by.params == Params(p=111)

    assert hash(Gen()) == hash(Gen(Params(p=111)))
    assert hash(Gen()) == hash(Gen(p=111))


def test_param_calls():
    """ " Test creation of `Primitive`s and `ExternalModule` calls with inline construction of their parameters."""

    Ext = h.ExternalModule(name="Ext", port_list=[], paramtype=dict)
    call = Ext(a=1, b="c", d=TabError)
    assert isinstance(call, h.ExternalModuleCall)
    assert call.params == dict(a=1, b="c", d=TabError)

    with pytest.raises(TypeError):
        Ext("not_a_dict_as_positional_arg")
    with pytest.raises(RuntimeError):
        Ext(dict(), b="d")

    @h.paramclass
    class P:
        reqd = h.Param(dtype=float, desc="reqd")
        a = h.Param(dtype=int, desc="a", default=111)
        b = h.Param(dtype=str, desc="b", default="c")

    Ext = h.ExternalModule(name="Ext", port_list=[], paramtype=P)
    call = Ext(reqd=3.14159, a=1, b="q")
    assert isinstance(call, h.ExternalModuleCall)
    assert call.params == P(reqd=3.14159, a=1, b="q")

    with pytest.raises(TypeError):
        Ext("not_a_P_as_positional_arg")
    with pytest.raises(RuntimeError):
        Ext(P(reqd=3.14159), a=12)

    from hdl21.primitives import R, ResistorParams

    call = R(r=11)
    assert isinstance(call, h.PrimitiveCall)
    assert call == R(R.Params(r=11))
    assert call == R(ResistorParams(r=11))

    with pytest.raises(TypeError):
        R("not_a_ResistorParams_as_positional_arg")
    with pytest.raises(RuntimeError):
        R(R.Params(r=11), z=12)


def test_bad_generators():
    """Test generator functions with bad call-signatures."""

    @h.paramclass
    class P:
        ...

    @h.generator  # A good, working generator!
    def good(p: P) -> h.Module:
        return h.Module()

    with pytest.raises(RuntimeError):
        # Bad 1st argument type
        @h.generator
        def bad(p: TabError) -> h.Module:
            return h.Module()

    with pytest.raises(RuntimeError):
        # Bad 2nd argument type
        @h.generator
        def bad(p: P, ctx: TabError) -> h.Module:
            return h.Module()

    with pytest.raises(RuntimeError):
        # Extra arg
        @h.generator
        def bad(p: P, something_else: int) -> h.Module:
            return h.Module()

    with pytest.raises(RuntimeError):
        # Bad return type
        @h.generator
        def bad(p: P) -> TabError:
            return h.Module()


def test_multi_portref_conns():
    """Test making several PortRef-based connections with a single assignment line."""

    @h.module
    class I:  # Inner Module
        a, b = h.Ports(2)

    @h.module
    class O:  # Outer, instantiating Module
        i1 = I()
        i2 = I()
        i3 = I()

        # The several connections
        i1.a = i2.a = i3.a

    # And do so again outside the class body
    O.i1.b = O.i2.b = O.i3.b

    h.elaborate(O)

    assert list(O.signals.keys()) == ["i3_a", "i3_b"]
    assert O.i1.conns["a"] is O.i2.conns["a"] is O.i3.conns["a"] is O.i3_a
    assert O.i1.conns["b"] is O.i2.conns["b"] is O.i3.conns["b"] is O.i3_b


def test_signal_usage():
    """Test that the `usage` property of `Signal`s is set correctly."""

    s = h.Signal()
    assert s.usage == h.Usage.SIGNAL
    p = h.Power()
    assert p.usage == h.Usage.POWER
    g = h.Ground()
    assert g.usage == h.Usage.GROUND
    c = h.Clock()
    assert c.usage == h.Usage.CLOCK

    s = h.Signals(2)
    assert all(s.usage == h.Usage.SIGNAL for s in s)
    p = h.Powers(3)
    assert all(p.usage == h.Usage.POWER for p in p)
    g = h.Grounds(4)
    assert all(g.usage == h.Usage.GROUND for g in g)
    c = h.Clocks(5)
    assert all(c.usage == h.Usage.CLOCK for c in c)


def test_properties():
    """Test adding `Properties` to signals, modules, instances, and the like."""

    s = h.Signal()

    s.props.set("foo", "bar")
    assert s.props.get("foo") == "bar"
    assert s.props["foo"] == "bar"

    s.props.set("foo", 4)
    assert s.props.get("foo") == 4
    assert s.props["foo"] == 4

    with pytest.raises(TypeError):
        s.props[3] = 4

    with pytest.raises(TypeError):
        s.props[TabError] = 11

    with pytest.raises(TypeError):
        s.props.set(key=None, val=5)

    with pytest.raises(TypeError):
        s.props.set(None, None)

    with pytest.raises(TypeError):
        s.props.get(None)

    m = h.Module(name="m")
    m.props["my_favorite_number"] = 42
    assert m.props["my_favorite_number"] == 42

    @h.module
    class M:
        ...

    M.props["your_favorite_number"] = 11
    assert M.props["your_favorite_number"] == 11

    mi = M()
    mi.props["abc"] = 123
    assert mi.props["abc"] == 123

    @h.bundle
    class B:
        ...

    B.props["their_favorite_number"] = 3
    assert B.props["their_favorite_number"] == 3

    bi = B()
    bi.props["xyz"] = None
    assert bi.props["xyz"] is None


def test_module_literals():
    """Test adding `Literal`s to modules."""

    @h.module
    class HasLit:
        x, y, z = 3 * h.Signal()

    HasLit.literals.append(h.Literal("mother"))
    HasLit.literals.append(h.Literal("father"))
    HasLit.literals.extend(
        [h.Literal("child"), h.Literal("uncle"), h.Literal("nephew")]
    )

    assert len(HasLit.signals) == 3
    assert HasLit.literals == [
        h.Literal(t) for t in ["mother", "father", "child", "uncle", "nephew"]
    ]


def test_signal_mult():
    """# Test using the multiplication operator to create multiple `Signal`s."""

    @h.module
    class M:
        a, b, c = 3 * h.Signal()
        i1, i2 = 2 * h.Input()
        o1, o2 = 2 * h.Output()
        io1, io2 = 2 * h.Inout()
        p1, p2, p3 = 3 * h.Port()

    assert M.a is not M.b
    assert M.a is not M.c
    assert M.b is not M.c
    assert M.i1 is not M.i2
    assert M.o1 is not M.o2
    assert M.io1 is not M.io2
    assert M.p1 is not M.p2

    assert len(M.ports) == 9
    assert len(M.signals) == 3


def test_bundle_mult():
    """# Test using the multiplication operator to create multiple `BundleInstance`s."""

    @h.bundle
    class B:
        i, q = 2 * h.Signal()

    @h.module
    class M:
        b1, b2 = 2 * B()  # <= here's the test!

    assert isinstance(M.b1, h.BundleInstance)
    assert isinstance(M.b2, h.BundleInstance)
    assert M.b1 is not M.b2

    h.elaborate(M)

    assert len(M.signals) == 4
    assert M.b1_i is not M.b2_i
    assert M.b1_q is not M.b2_q


def test_set_bad_attrs():
    """Test the "only allow setting known attributes" feature of a few types."""

    @h.bundle
    class B:
        s = h.Signal()

    # Create a `BundleInstance`
    b = B()

    # And assert we can't assign stuff to it
    with pytest.raises(RuntimeError):
        b.q = 5

    # Create a `BundleRef` into that instance
    ref = b.some_signal

    # And assert we can't assign stuff to it
    with pytest.raises(RuntimeError):
        ref.whatever = 6

    # Create an `AnonymousBundle`
    bu = h.bundlize(a=h.Signal())

    # And assert we can't assign stuff to it
    with pytest.raises(RuntimeError):
        bu.q = 5

    # Create a `Concat`
    c = h.Concat(h.Signal(), h.Signal())

    # And assert we can't assign stuff to it
    with pytest.raises(RuntimeError):
        c.xyz = TabError


def test_generator_function_combo():
    """Test instantiating hierarchies that combine `@h.generator`s and regular old functions.
    Inspired #192 https://github.com/dan-fritchman/Hdl21/issues/192."""

    @h.generator
    def Gen(_: h.HasNoParams) -> h.Module:
        @h.module
        class HasDiff:
            d = h.Diff(port=True)

        return HasDiff

    def NonGen() -> h.Module:
        @h.module
        class HasTwoHasDiffs:
            d = h.Diff(port=False)
            a = Gen()(d=d)
            b = Gen()(d=d)

        return HasTwoHasDiffs

    h.elaborate(NonGen())
    h.elaborate(NonGen())


def test_how_elab_caching_works():
    """Not exactly a test of Hdl21, but of the mechanism used by class-level elaborator caching.
    More or less the fix for #192 https://github.com/dan-fritchman/Hdl21/issues/192."""

    from typing import Optional, List
    from dataclasses import dataclass, field

    @dataclass
    class Cache:
        ls: List[int] = field(default_factory=list)

    class Base:
        CLASS_CACHE: Optional[Cache] = None

        def __init_subclass__(cls) -> None:
            # Create a new `Cache` for each subclass
            cls.CLASS_CACHE = Cache()

        def do_something(self) -> None:
            # Modify the class-level cache in an instance method
            self.CLASS_CACHE.ls.append(1)

    class Sub1(Base):
        pass

    class Sub2(Base):
        pass

    sub1 = Sub1()
    sub2 = Sub2()

    assert Base.CLASS_CACHE is None
    assert Sub1.CLASS_CACHE is not None
    assert Sub2.CLASS_CACHE is not None
    assert Sub1.CLASS_CACHE is not Sub2.CLASS_CACHE
    assert sub1.CLASS_CACHE is not None
    assert sub2.CLASS_CACHE is not None
    assert sub1.CLASS_CACHE is not sub2.CLASS_CACHE

    sub1.do_something()
    sub1.do_something()

    assert Base.CLASS_CACHE is None
    assert sub1.CLASS_CACHE.ls == [1, 1]
    assert sub2.CLASS_CACHE.ls == []

    sub2.do_something()
    sub2.do_something()
    sub2.do_something()

    assert Base.CLASS_CACHE is None
    assert sub1.CLASS_CACHE.ls == [1, 1]
    assert sub2.CLASS_CACHE.ls == [1, 1, 1]

    sub1b = Sub1()
    sub2b = Sub2()

    assert Base.CLASS_CACHE is None
    assert sub1b.CLASS_CACHE.ls == [1, 1]
    assert sub2b.CLASS_CACHE.ls == [1, 1, 1]
    assert sub1.CLASS_CACHE is sub1b.CLASS_CACHE
    assert sub2.CLASS_CACHE is sub2b.CLASS_CACHE


def test_module_circular_dependency():
    """Test attempting to elaborate an invalid circular dependency loop between `Module`s."""

    # Create two modules which errantly instantiate each other
    a = h.Module(name="a")
    b = h.Module(name="b")
    a.add(h.Instance(of=b, name="ib"))
    b.add(h.Instance(of=a, name="ia"))

    # Test that attempting to elaborate either throws the circular-dependency error

    with pytest.raises(RuntimeError) as einfo:
        h.elaborate(a)
    assert "Invalid self referencing/ circular dependency" in str(einfo.value)

    with pytest.raises(RuntimeError) as einfo:
        h.elaborate(b)
    assert "Invalid self referencing/ circular dependency" in str(einfo.value)


def test_generator_circular_dependency():
    """Test attempting to elaborate an invalid circular dependency loop between `Generator`s."""

    # Create two generators which errantly instantiate each other
    @h.generator
    def f(_: h.HasNoParams) -> h.Module:
        m = h.Module()
        m.g = g()()
        return m

    @h.generator
    def g(_: h.HasNoParams) -> h.Module:
        m = h.Module()
        m.f = f()()
        return m

    # Test that attempting to elaborate either throws the circular-dependency error

    with pytest.raises(RuntimeError) as einfo:
        h.elaborate(f())
    assert "Invalid self referencing/ circular dependency" in str(einfo.value)

    with pytest.raises(RuntimeError) as einfo:
        h.elaborate(g())
    assert "Invalid self referencing/ circular dependency" in str(einfo.value)


def test_generator_call_as_param():
    """Test using a generated module as a parameter.
    Inspired by #93 https://github.com/dan-fritchman/Hdl21/issues/93."""

    from hdl21.generators import MosStack as Hdl21Mos

    @h.generator
    def Pmos(_: h.HasNoParams) -> h.Module:
        return Hdl21Mos()  # Wrap the built-in generator

    @h.paramclass
    class Params:
        unit = h.Param(dtype=h.Instantiable, desc="the unit pmos")

    @h.generator
    def PmosPair(params: Params) -> h.Module:
        @h.module
        class PmosPair:
            x = h.Signal()
            pair = h.Pair(params.unit)(d=x, g=x, s=x, b=x)

        return PmosPair

    h.elaborate(PmosPair(Params(Pmos())))


def test_generator_caching():
    """Test generators with and without caching.
    https://github.com/dan-fritchman/Hdl21/issues/198"""

    @h.generator
    def G(_: h.HasNoParams) -> h.Module:
        return h.Module()

    g1 = G()
    g2 = G()
    assert g1 is g2

    @h.generator(enable_cache=False)
    def N(_: h.HasNoParams) -> h.Module:
        return h.Module()

    n1 = N()
    n2 = N()
    assert n1 is not n2

    # Clear the cache and re-run em a few times
    h.generator.cache.reset()

    g3 = G()
    g4 = G()
    assert g3 is g4
    assert g3 is not g2

    n3 = N()
    n4 = N()
    assert n3 is not n4
    assert n3 is not n2
    assert n3 is not n1



"""
Hdl21 Parameters and Param-Classes 
"""

# Std-Lib Imports
import dataclasses, inspect, json, hashlib
from typing import Optional, Any, Type, TypeVar, Dict

# PyPi Imports
import pydantic

# Local Imports
from .default import Default
from .datatype import AllowArbConfig, pydantic_json_encoder

T = TypeVar("T")


def paramclass(cls: Type[T]) -> Type[T]:
    """# Parameter-Class Creation Decorator

    Transforms a class-definition full of Params into a type-validated dataclass,
    with methods for default value and description-dictionary retrieval.

    Hdl21's `paramclass`es are immutable, strongly-typed data-storage structures.
    They are defined through a syntax similar to `@dataclass`, but using the `Param`
    constructor, and assignment rather than type annotation.

    ```python
    @paramclass
    class C:
        reqd = Param(dtype=int, desc="A Required Parameter")
        optn = Param(dtype=int, desc="An Optional Parameter", default=11)
    ```

    `Param`s each have required datatype (`dtype`) and description (`desc`) fields,
    and optional default values.
    Each `paramclass` constructor can be called with ordered arguments,
    in the order defined in the `paramclass`, or with named arguments.
    Named arguments are highly recommended for more than a single parameter.
    Note Python's function-argument ordering requirements also dictate
    that all `paramclass` required-arguments be declared *before* any optional arguments.
    This also reinforces good practice for communicating which parameters are required.

    Each `paramclass` comes with class-methods `descriptions` and `defaults`,
    which return dictionaries of the parameter names to descriptions and
    names to default values (for those with defaults), respectively.

    Requirements of the input `cls`:
    * *All* non-Python-internal fields must be of type `Param`
    * Inheritance is not supported
    """

    if not inspect.isclass(cls):
        msg = f"Invalid @hdl21.paramclass `{cls}`. Apply the `@paramclass` decorator to classes."
        raise RuntimeError(msg)
    if cls.__bases__ != (object,):
        raise RuntimeError(f"Invalid @hdl21.paramclass inheriting from {cls.__bases__}")

    protected_names = ["descriptions", "defaults"]
    dunders = dict()  # Double-under attributes; ignored
    params = dict()  # Parameters; used to create dataclass fields

    # Take a lap through the class dictionary, type-check everything and grab Params
    # FIXME: look for, and alert users about, the error writing type annotations rather than equality.
    # (Or, we could move to type annotations...)
    for key, val in cls.__dict__.items():
        if key in protected_names:
            raise RuntimeError(f"Invalid field name {key} in paramclass {cls}")
        elif key.startswith("__"):
            dunders[key] = val
        elif isinstance(val, Param):
            params[key] = val
        else:
            msg = f"Invalid class-attribute {key} in paramclass {cls}. All attributes should be `hdl21.Param`s."
            raise RuntimeError(msg)

    # Set up type annotations for the dataclass-to-be
    # There is quite a bit of invaluable content on this here:
    # https://docs.python.org/3/howto/annotations.html
    # Especially for Python versions before 3.10.
    # Some of our constraints - notably the lack of inheritance, checked above -
    # make that advice somewhat easier to follow;
    # nonetheless we are definitely not following it in full.

    # Notably, this weird bit right here is the "howto/annotations" recommendation -
    # Look only in the class `__dict__`, and be prepared for `__annotations__` to be missing.
    annotations = cls.__dict__.get("__annotations__", None)
    if annotations is None:
        annotations = cls.__annotations__ = dict()

    # Translate the Params into dataclass-compatible annotations
    for name, param in params.items():
        # Set the type-annotation
        annotations[name] = param.dtype

        # Handle the default value, which is the class-attribute-value on dataclasses
        if param.default is not Default and param.default_factory is not Default:
            msg = f"Invalid Param {param} in paramclass {cls}. Only one of `default` or `default_factory` may be specified."
            raise RuntimeError(msg)

        elif param.default is not Default:
            setattr(cls, name, param.default)

        elif param.default_factory is not Default:
            field = dataclasses.field(default_factory=param.default_factory)
            setattr(cls, name, field)

        else:  # Otherwise there is no default value, this is a required parameter.
            # Remove the `Param` object from the class namespace
            delattr(cls, name)

    # Add a few helpers to the class namespace
    cls.__params__ = params
    cls.__paramclass__ = True
    cls.descriptions = classmethod(descriptions)
    cls.default_instance = classmethod(default_instance)
    cls.default_dict = classmethod(default_dict)
    cls.defaults = classmethod(defaults)

    # Pass this through the pydantic dataclass-decorator-function
    cls = pydantic.dataclasses.dataclass(cls, config=AllowArbConfig, frozen=True)

    # Pydantic seems to want to add this one *after* class-creation
    def _brick_subclassing_(cls, *_, **__):
        msg = f"Error: attempt to sub-class `hdl21.paramclass` {cls} is not supported"
        raise RuntimeError(msg)

    cls.__init_subclass__ = classmethod(_brick_subclassing_)
    # And don't forget to return it!
    return cls


def descriptions(cls: Type) -> Dict[str, str]:
    """# Get a dictionary of parameter names to descriptions for `cls`
    Available both as a free function and as a `classmethod` of each `paramclass`.
    Raises a `RuntimeError` if `cls` is not a param-class."""

    if not isparamclass(cls):
        raise RuntimeError(f"Invalid @hdl21.paramclass {cls}")

    return {k: v.desc for k, v in cls.__params__.items()}


def default_instance(cls: Type[T]) -> T:
    """# Default Instance
    Create the default instance of param-class `cls`.
    Available both as a free function and as a `classmethod` of each `paramclass`.
    Throws an exception if `cls` is not a param-class, or is not default-constructible.
    """
    if not isparamclass(cls):
        raise RuntimeError(f"Invalid @hdl21.paramclass {cls}")
    return cls()


def default_dict(cls: Type[T]) -> Dict[str, Any]:
    """# Default Values Dictionary

    Get a dictionary of parameter names to default values for param-class `cls`.
    Available both as a free function and as a `classmethod` of each `paramclass`.
    Raises a `RuntimeError` if `cls` is not a param-class.
    *Does not* require that `cls` is default constructible.
    Returns the default values for whatever parameters have them."""

    if not isparamclass(cls):
        raise RuntimeError(f"Invalid @hdl21.paramclass {cls}")

    rv: Dict[str, Any] = {}
    for k, v in cls.__params__.items():
        if v.default is not Default:
            rv[k] = v.default
        if v.default_factory is not Default:
            rv[k] = v.default_factory()
    return rv


# Retain the original name `defaults` for the dictionary version
defaults = default_dict


def isparamclass(cls: type) -> bool:
    """Boolean indication of whether `cls` has been `@paramclass`-decorated"""
    return getattr(cls, "__paramclass__", False)


def hasparams(cls: type) -> bool:
    """Boolean indication of whether `cls` has a nonzero number of parameters.
    `HasNoParams` is a prominent, built-in example of a `paramclass` with no parameters.
    Throws an exception is `cls` is not a paramclass."""
    if not isparamclass(cls):
        raise RuntimeError(f"Invalid @hdl21.paramclass {cls}")
    return len(cls.__params__) > 0


@pydantic.dataclasses.dataclass(config=AllowArbConfig, frozen=True)
class Param:
    """# Parameter Declaration"""

    dtype: Any  # Datatype. Required
    desc: str  # Description. Required
    default: Optional[Any] = Default  # Default Value. Optional
    default_factory: Optional[Any] = Default  # Default Call-Factory. Optional


def _unique_name(params: Any) -> str:
    """Create a unique name for parameter-class instance `params`"""
    if not isparamclass(params):
        raise RuntimeError(f"Invalid parameter-class instance {params}")

    # Determine whether *all* fields of `params` are scalar values: strings, numbers, and options thereof
    scalars = [
        str,
        int,
        float,
        type(None),
        Optional[str],
        Optional[int],
        Optional[float],
    ]
    # Boolean indication of whether *all* param-datatypes are from among these
    all_scalar = all([param.dtype in scalars for param in params.__params__.values()])

    # If all params are scalars, create a readable string of their values
    if all_scalar:
        # Format: `pname1=pval1 pname2=pval2 pname3=pval3`
        keys = params.__params__.keys()
        name = " ".join(f"{k}={str(getattr(params, k))}" for k in keys)

        # These names must also be limited in length, for sake of our favorite output formats.
        # If the generated name is too long, use the hashing method below instead
        if len(name) < 128:  # Probably(?) a reasonable length limit
            return name

    # Non-scalar cases generally include nested `@paramclasses` or sequences,
    # or surpass the length-limits above. We serialize and hash them.
    # Preferably serialize as JSON
    jsonstr = json.dumps(params, indent=4, default=hdl21_naming_encoder)
    data = bytes(jsonstr, encoding="utf-8")

    # If JSON encoding fails, we *could* use pickle instead.
    # Disabled for now, because pickle is not guaranteed to be deterministic.
    # Note JSON is preferable for its run-to-run stability,
    # where pickle can generally serialize more types.
    # data = pickle.dumps(params)

    # Take that data and hash it
    # Note the "not used for security" option ensures consistent hashing between runs/ Python-processes
    h = hashlib.new("md5", usedforsecurity=False)
    h.update(data)
    # And return the (hex) digest as our unique name
    return h.hexdigest()


def hdl21_naming_encoder(obj: Any) -> Any:
    """JSON encoder for naming of Hdl21 parameter-values.

    "Extends" `pydantic.json.pydantic_encoder` by first checking for
    each of the non-serializable Hdl21 types (`Module`, `Instance`, `Generator`, etc.),
    then hands everything else off to `pydantic.json.pydantic_encoder`.

    Note this *does not fully serialize `Module`s and the like -
    see `hdl21.to_proto` for this. This JSON-ization is just good enough
    to enable unique naming of Hdl-object-valued parameters."""

    from .module import Module
    from .qualname import qualname as module_qualname
    from .external_module import ExternalModule, ExternalModuleCall
    from .instance import Instance
    from .generator import Generator
    from .primitives import Primitive, PrimitiveCall

    if isinstance(obj, (Instance,)):
        # Not supported as parameters
        raise RuntimeError(f"Invalid `hdl21.paramclass` field {obj}")

    if isinstance(obj, (Module, ExternalModule, Generator)):
        # Use qualified class names/paths
        return module_qualname(obj)

    if isinstance(obj, (Primitive, PrimitiveCall)):
        # Primitives use their `name` attribute/ property directly
        # FIXME: this should really mix in the parameters as well, as `ExternalModuleCall` does.
        return obj.name

    if isinstance(obj, ExternalModuleCall):
        # Mix the qualified class names/paths with the parameters
        return module_qualname(obj.module) + _unique_name(obj.params)

    # Dataclasses also require custom handling, as the default encoder deep-copies them,
    # often invoking methods not supported on several Hdl21 types.
    # Convert to (shallow) dictionaries instead.
    if dataclasses.is_dataclass(obj):
        return {f.name: getattr(obj, f.name) for f in dataclasses.fields(obj)}

    # Not an Hdl21 type. Hand off to pydantic.
    return pydantic_json_encoder(obj)


# Shortcut for parameter-less generators.
# Defines the empty-value `@paramclass`.
@paramclass
class HasNoParams:
    """# HasNoParams
    A built-in `hdl21.paramclass` for generators that require no parameters."""

    ...  # Empty Contents


# And an instance of it
NoParams = HasNoParams()

__all__ = ["paramclass", "Param", "HasNoParams", "NoParams"]



"""
# Hdl21 PDK Sample 

An entirely fictitious process-technology, 
written solely as a demonstration of an `hdl21.pdk` plug-in package. 
"""

from pathlib import Path
from typing import Optional
from hdl21.pdk import register

# Grab our primary PDK-definition module
from . import pdk
from .pdk import *


# And register as a PDK module
register(pdk)

# Provide a `resources` path to our non-Python data
# This enables usage along the lines of
# `sample_pdk.resources / "a_file.txt"
resources = Path(__file__).parent / "resources"

# The optional external-data installation.
# Set by an instantiator of `Install`, if available.
# Unlike most PDK pacakges, this sample provides a default `Install`,
# as it has a very small amount of non-Python data,
# which is tracked in this repository.
install: Optional[Install] = Install(models=resources / "models.sp")



# Std-Lib imports
from typing import Set, Union, Optional

# Local imports
from .datatype import datatype, AllowArbConfig
from .connect import connectable
from .sliceable import sliceable
from .concat import concatable
from .instance import _Instance


@concatable
@sliceable
@connectable
@datatype(config=AllowArbConfig)
class PortRef:
    """
    # PortRef
    Reference to a Port
    Created from a combination of a parent `Instance` and a port-name.
    """

    # Core fields: the Instance and port name
    inst: _Instance
    portname: str

    def __post_init__(self):
        # Inner management data
        self._connected_ports: Set[PortRef] = set()
        self.resolved: Union[None, "Signal", "BundleInstance"] = None
        self._slices: Set["Slice"] = set()
        self._concats: Set["Concat"] = set()
        self._width: Optional[int] = None

    def __eq__(self, other) -> bool:
        """Port-reference equality requires *identity* between instances
        (and of course equality of port-name)."""
        if not isinstance(other, PortRef):
            return False
        return self.inst is other.inst and self.portname == other.portname

    def __hash__(self):
        """Hash references as the tuple of their instance-address and name"""
        return hash((id(self.inst), self.portname))



""" 
# Elaboration Passes

Recursive walkers of Hdl21 hierarchies which perform the mechanics of elaboration. 
"""

# Import the abstract base class for typing guards
from .base import ElabPass

# And all the built-in passes
from .inst_bundles import InstBundleElabPass
from .slices import SliceResolver
from .orphanage import Orphanage
from .arrays import ArrayFlattener
from .portrefs import ResolvePortRefs
from .conntypes import ConnTypes
from .flatten_bundles import BundleFlattener
from .mark_modules import MarkModules
